#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

require 'delay_evaluation'

module SidfmControllerSupport

  # Exceptionをログ出力
  def log_exception(e)
    logger.error(e.message)
    if e.backtrace then
      e.backtrace.each do |l|
        logger.error(l)
      end
    end
  end

  # sidebar等、全ての画面で必要なデータを取得する。
  # 全て遅延評価で入れること
  def create_common_data
    @common = {}
    # システムカテゴリ一覧。ホストが紐づけられているもののみ。
    @common[:system_categories] = DelayEvaluation.new do
      SidfmdCategory.system.with_hosts
    end
    # システムカテゴリ数。ホストが紐づけられているもののみ。
    @common[:system_categories_size] = DelayEvaluation.new do
      @common[:system_categories].v.size
    end
    # システムカテゴリ以外のカテゴリ一覧。
    # ホストが紐づけられているもののみ。
    @common[:not_system_categories] = DelayEvaluation.new do
      if @sidfm_roles and !@sidfm_roles.include?(:admin) then
        SidfmdCategory.not_system.with_hosts.where(:id => User.current.categories.select(:id))
      else
        SidfmdCategory.not_system.with_hosts
      end
    end
    # システムカテゴリ以外のカテゴリ数。
    # ホストが紐づけられているもののみ。
    @common[:not_system_categories_size] = DelayEvaluation.new do
      @common[:not_system_categories].v.size
    end
    # 有効な（削除されていない）ホストのOSとバージョンの情報
    @common[:dist_versions] = DelayEvaluation.new do
      SidfmdQueueDist.get_dist_versions
    end
    @common[:dist_versions_size] = DelayEvaluation.new do
      @common[:dist_versions].v.size
    end
    # 以下の状態の対策予定日付を返す。
    # ・有効な（削除されていない）ホスト
    # ・未確認、対策中のステータス
    @common[:due_dates] = DelayEvaluation.new do
      SidfmdIssue.get_due_dates
    end
    # :due_datesの数
    @common[:due_dates_size] = DelayEvaluation.new do
      @common[:due_dates].v.size
    end
    # 危険度の一覧。高い順。
    @common[:risks] = DelayEvaluation.new do
      SidfmdRisk.all.order(:id => :desc)
    end
    # 対応状況の一覧。id順（未対策->対処済)
    @common[:statuses] = DelayEvaluation.new do
      IssueStatus.all.order(:id)
    end
  end

  # CVE IDの正規化
  def normalize_cve(txt)
    ret = nil
    if /CVE-?(\d{4})-?(\d+)/i =~ txt then
      ret = "CVE-#{Regexp.last_match(1)}-#{Regexp.last_match(2)}"
    end
    ret
  end

  def display_percent(v)
    v ? "#{sprintf('%.1f', v)}%" : '---'
  end

  # 対策予定日設定済表示用
  def display_due_dates_percent(due_dates)
    return '---' unless due_dates
    f_cnt = due_dates.size
    cnt = due_dates.inject(0) {|r, x| r += 1 if (x and x.size > 0); r }
    display_percent(cnt.to_f / f_cnt.to_f * 100.0)
  end

  # 対応状況表示用
  def display_finished_percent(status_ids)
    return '---' unless status_ids
    f_cnt = status_ids.size
    cnt = status_ids.inject(0) {|r, x|
      r += 1 if SidfmdIssue::STATUS_FINISHED.include?(x)
      r
    }
    display_percent(cnt.to_f / f_cnt.to_f * 100.0)
  end

  PROJECT_SIDFM = '脆弱性情報管理'.freeze

  # SIDfmVMのプロジェクトを取得
  def get_sidfm_project
    unless @sidfm_project then
      @sidfm_project = Project.where(:name => PROJECT_SIDFM).first
    end
    @sidfm_project
  end

  ROLE_SIDFM_ADM = 'SIDfmVM管理者'.freeze
  ROLE_SIDFM_MNG = 'SIDfmVM報告者'.freeze
  ROLE_SIDFM_USE = 'SIDfmVM利用者'.freeze
  ROLE_SIDFM_USR = 'SIDfmVM閲覧者'.freeze
  ROLES_SIDFM = [ ROLE_SIDFM_ADM, ROLE_SIDFM_MNG, ROLE_SIDFM_USE, ROLE_SIDFM_USR ].freeze

  # SIDfmVMのカレントユーザを取得
  def get_sidfm_user
    unless @sidfm_user then
      @sidfm_user = User.current
      if @sidfm_user then
        if @sidfm_user.anonymous? then
          @sidfm_user = nil
        else
          roles = @sidfm_user.roles_for_project(get_sidfm_project)
          @sidfm_roles = []
          roles.each do |role|
            perms = role.permissions
            if perms then
              @sidfm_roles << :admin if perms.include?(:admin_sidfm)
              @sidfm_roles << :use if perms.include?(:use_sidfm)
              @sidfm_roles << :view if perms.include?(:view_sidfm)
              @sidfm_roles << :user_admin if perms.include?(:user_admin_sidfm)
              @sidfm_roles << :host_admin if perms.include?(:host_admin_sidfm)
              @sidfm_roles << :host_limited_admin if perms.include?(:host_limited_admin_sidfm)
              @sidfm_roles << :rule_admin if perms.include?(:rule_admin_sidfm)

            end
          end
          @sidfm_roles.uniq!
        end
      end
    end
    unless @sidfm_roles then
      @sidfm_roles = []
    end
    @sidfm_user
  end

  # 現在アクセス中のユーザの権限の取得
  def get_sidfm_roles
    unless @sidfm_roles then
      get_sidfm_user
    end
    @sidfm_roles
  end

  # ライセンス登録の確認
  def check_license
    if Sidfm.sidfmdlm_license_key.nil? then
      redirect_to :controller => :sidfm_licenses, :action => :license, :notice => 'ライセンス未登録です。' and return
    end
  end

  #Sessionのupdate_atを更新
  def update_session_updated_at
    current_session = Session.find_by(:session_id => request.session_options[:id])
    if current_session then
      current_session.update_attribute(:updated_at, Time.current)
    end
  end

  #importで使うsessionのリセット
  def clear_import_session
    #共通
    session[:import_notices] = nil
    session[:import_actions] = nil
    session[:import_check_data] = nil #データの重複チェック
    session[:clear_session] = nil #indexページなどで、これがtrueならclear_import_sessionを行う
    #ホスト設定画面のnew_import
    session[:import_errs] = nil
    session[:lmhosts] = nil
    session[:is_new_import] = nil
    session[:hosts_info_file_lines] = nil
    session[:host_opts_info_files] = nil
    session[:new_host_opts_info_files] = nil
    session[:host_opts_info] = nil
    session[:new_host_opts_info] = nil
    session[:line_num] = nil
    session[:new_host_ids] = nil
    session[:import_hostnames] = nil
    #ローカルプロダクトのimport
    session[:import_lproducts] = nil
    session[:import_lversions] = nil
    #ローカル脆弱性のimport
    session[:import_local_vul_files] = nil
    #カテゴリ設定画面のimport
    session[:import_categories] = nil
    #ユーザ管理画面のimport
    session[:import_users] = nil
  end
  
  #exportで使うsessionのリセット
  def clear_export_session
    #ホスト設定画面のnew_export
    session[:export_hosts] = nil
    session[:export_hosts_lines] = nil
    session[:export_opts_lines] = nil
    #ローカルプロダクトのexport
    session[:export_lproducts] = nil
    session[:export_lproduct_lines] = nil
    #ローカル脆弱性のexport
    session[:export_local_vuls] = nil
    session[:export_local_vul_lines] = nil
    #履歴画面のexport ※1ファイルダウンロード場合はここの値を使う
    session[:export_data] = nil
  end

  # アクセス失敗時にユーザ名をロギングする
  # 返り値は引数のmsgをそのまま返す。
  def af_log_user(msg)
    logger.info("Access Failed: #{User.current}, #{msg}")
    msg
  end

  def get_view_state
    @view_state = {}
    @view_state[:idx] = (cookies[:state_idx] || 'sri').to_sym
  end

  def put_view_state
    cookies[:state_idx] = @view_state[:idx].to_s
  end

  def get_vm_version 
    if FileTest.exist?('/sidfmvm-version') then
      @vm_version = File.read('/sidfmvm-version')
    else
      @vm_version = nil
    end
  end

  def get_summary_table_lines(page_name)
    @summary_table_lines = cookies['sidfm_' + page_name + '_summary_num']
  end

  def get_summary_prior_category
    #今のところ'custom'か'system'のいずれか
    @summary_prior_category = cookies['sidfm_summary_prior_category'] || 'system'
  end

=begin
  #def create_vm_export_json_file_data(data, block)
  #  str = ''
  #  first = true
  #  param_info = JSON.generate(data[:param_info])
  #  str += param_info.chop! + ',"content":['#ハッシュの中にcontentのキーでデータ入れるため、param_info.chop!で末尾(})を切る
  #  data[:content].each do |row|
  #    row = JSON.generate(block.call(row))
  #    if !first then
  #      row = ',' + row
  #    else
  #      first = false
  #    end
  #    str += row
  #  end
  #  str += (']}')
  #  str
  #end

  #jsonファイル送信前にクエリからデータをファイルで生成
  def create_vm_export_json_file(data, block)
    file = nil
    file_path = nil
    first = true
    param_info = JSON.generate(data[:param_info])
    Tempfile.open('temp'){|tp|
      tp.write(param_info.chop! + ',"content":[') #ハッシュの中にcontentのキーでデータ入れるため、param_info.chop!で末尾(})を切る
      data[:content].each do |row|
        str = JSON.generate(block.call(row))
        if !first then
          str = ',' +  str
        else
          first = false
        end
        tp.write(str)
      end
      tp.write(']}')
      file_path = tp.path
      file = tp
    }
    return { :file => file, :file_path => file_path }
  end

  #jsonファイルをstreamで送信
  STREAM_STRING_SIZE = 500000
  def export_summary_json_file(file, file_path, file_name)
    begin
      response.headers['Cache-Control'] = 'no-cache'
      response.headers['Content-Type'] = 'text/json'
      response.headers['Content-Disposition'] = "attachment; filename=#{file_name}"
      #response.headers['Content-Length'] = file.size.to_s
        
      File.open(file_path) do |f|
        f.each(STREAM_STRING_SIZE) do |str|
          response.stream.write(str)
          GC.start
        end
      end
    ensure
      file.unlink
      response.stream.close
    end
  end

  #csvファイルをstreamで送信
  #rowにはエンコードしていて且つクエリを発行しないcsvデータ(配列)を渡す
  CSV_START_GC_SIZE = 500
  def export_summary_tsv_file(file_name, rows)
    begin
      csv_bom = "\xff\xfe".force_encoding('UTF-16LE')
      response.headers['Cache-Control'] = 'no-cache'
      response.headers['Content-Type'] = 'text/csv'
      response.headers['Content-Disposition'] = "attachment; filename=#{file_name}"
      cnt = 0
      rows.each do |row|
        response.stream.write(row)
        cnt += 1
        if cnt == CSV_START_GC_SIZE then
          GC.start
          cnt = 0
        end
      end
    ensure
      response.stream.close
    end
  end
=end

  DOWNLOAD_READ_SIZE = 500
  #サマリー表のJSONファイルダウンロード
  #process_idsは、単純にsliceで処理できない際に、処理するidを指定する時に使う
  def init_json_summary_table_progress_download(**opts)
    session[:export_data] = {}
    session[:export_data][:order] = { :by => opts[:order_col_name], :type => opts[:order_type] }
    session[:export_data][:filters] = opts[:filters]

    session[:export_data][:category] = opts[:category] if opts[:category]
    session[:export_data][:idx] = opts[:idx] if opts[:idx]
    session[:export_data][:host] = opts[:host_detail_info] if opts[:host_detail_info] #ホスト詳細のみ

    session[:export_data][:content] = []
    #ダウンロード中(読み込み中)にデータが書き換えられても全てデータが出力されるように、idでデータを順次取得していく。
    render :layout => false, :json => { :err_message => '', :ids => opts[:ids], :max_download_size => opts[:max_download_size], :executed_size => 0, :process_ids => opts[:process_ids] } and return
  end

  def send_summary_table_json_file(filename)
    begin
      send_data(session[:export_data].to_json, :filename => filename, :disposition => 'attachment')
      cookies['export_flg'] = true
    ensure
      session[:export_data] = nil
    end
  end

  #サマリー表のTSVファイルダウンロード
  #process_idsは、単純にsliceで処理できない際に、処理するidを指定する時に使う
  def init_tsv_summary_table_progress_download(ids, header, max_download_size, process_ids = nil)
    session[:export_data] = {}
    session[:export_data][:content] = []
    session[:export_data][:content] << header
    #ダウンロード中(読み込み中)にデータが書き換えられても全てデータが出力されるように、idでデータを順次取得していく。
    render :layout => false, :json => { :err_message => '', :ids => ids, :max_download_size => max_download_size, :executed_size => 0, :process_ids => process_ids } and return
  end
  
  def send_summary_table_tsv_file(filename)
    csv_bom = "\xff\xfe".force_encoding('UTF-16LE').freeze
    begin
      csv = CSV.generate("#{csv_bom}".force_encoding('UTF-16LE'), :col_sep => "\t".encode('UTF-16LE')) do |o|
        session[:export_data][:content].each do |row|
          o << row.map{|item| 
            begin
              item = item.encode('UTF-16LE')
            rescue Exception => e
              #万が一nilや数値が入っていた場合は、internal errorを防ぐために空文字にする。
              logger.info(e)
              item = ''.encode('UTF-16LE')
            end
            item
          }
        end
      end
      send_data(csv, :filename => filename, :disposition => 'attachment')
      cookies['export_flg'] = true
    ensure
      session[:export_data] = nil
    end
  end

  #サマリー表のplaybookファイルダウンロード
  def send_summary_table_playbook_file(filename)
    begin
      playbook = GenerateAnsiblePlaybook.new(session[:export_data].to_json).generate
      if playbook != 'error' && playbook != 'no_packages' then
        send_data(playbook, :filename => filename, :disposition => 'attachment')
        cookies['export_flg'] = true
      end
    ensure
      session[:export_data] = nil
    end
  end

  #サマリー表のJSON/TSVファイルダウンロード 共通で使うメソッド
  #サマリー表のcontentを生成
  #一度に処理する数はread_sizeに値を渡して調整可能
  def json_tsv_summary_table_progress_download(ids, block, executed_size, read_size = DOWNLOAD_READ_SIZE)
    ids.shift(read_size).each do |id|
      item = block.call(id)
      session[:export_data][:content] << item if item
      executed_size += 1
    end
    render :layout => false, :json => { :err_message => '', :ids => ids, :executed_size => executed_size } and return
  end

  CACHE_OPTIONS = { :expires_in => 1.hour, :race_condition_ttl => 1.minute }
  
  #履歴通知機能で表示する履歴の期間
  NOTIFY_HISTORY_DATE = (Date.today - 7).to_time

  #履歴のコメント削除で差し替える内容
  HISTORY_COMMENT_DELETED_NOTE = 'コメントは削除されました。'

  OPENID_CONNECT_AUTH_NAME = 'OpenIDConnect'

  # TSV出力時のカテゴリのエスケープ処理
  def escape_category_tsv(s)
    "\"#{s ? s.gsub(/\"/, '""') : '' }\""
  end

  # フィルタでユーザのアクセス範囲外のデータを削除する
  # 保存フィルタ反映時に、ユーザのアクセス範囲外のデータしかないとカテゴリ等の検索で
  # 表示される内容が空なのに検索状態となってしまうのを防ぐため
  # user_host_osnamesにはユーザのアクセス範囲内のホストのOSを渡す
  def filter_filters(filters:, user_host_osnames: [], role_admin:, user_category_ids:)
    filters = filters.deep_dup
    if role_admin then
      return filters #管理者は制限ないためフィルタ不要
    end
    if filters.has_key?(:categories) then
      if filters[:categories][:ids] && filters[:categories][:ids].any? then
        filters[:categories][:ids] = filters[:categories][:ids].map(&:to_i) & (SidfmdCategory.where(:is_system => true).pluck(:id) + user_category_ids)
        if filters[:categories][:ids].size == 0 then
          if filters[:categories][:is_and] then
            filters[:categories][:is_and] = false
          end
          if filters[:categories][:is_except] then
            filters[:categories][:is_except] = false
          end
        end
      end
    end
    if filters.has_key?(:ctags) then
      if filters[:ctags][:ids] && filters[:ctags][:ids].any? then
        filters[:ctags][:ids] = filters[:ctags][:ids].map(&:to_i) & SidfmdCtag.where("sidfmd_category_id IS NULL OR sidfmd_category_id IN (?)", user_category_ids).pluck(:id)
      end
    end
    if filters.has_key?(:os) then
      if filters[:os].size > 0 then
        if !user_host_osnames.include?(filters[:os]) then
          filters[:os] = ''
        end
      end
    end
    return filters
  end

  #Hashのデータの中身の型チェック
  #1つでも型が異なるデータがある/キーがないデータがある場合はfalseを返す
  #ただし、データの型については、valueがnilの場合はpassとする。
  def compare_hash_type(hash1, hash2)
    is_same_type = true
    if (hash1.keys - hash2.keys).size > 0 || (hash2.keys - hash1.keys).size > 0 then
      is_same_type = false
    else
      hash1.each do |key, val|
        if val == nil or hash2[key] == nil then
          next
        end
        if !val.is_a?(hash2[key].class) then
          if val == true || val == false then
            is_same_type = false if hash2[key] != true && hash2[key] != false
          else
            is_same_type = false
          end
        else
          if val.is_a?(Hash) then
            temp_is_same_type = compare_hash_type(val, hash2[key])
            unless temp_is_same_type then
              is_same_type = false
            end
          elsif val.is_a?(Array) then
            temp_is_same_type = compare_array_type(val, hash2[key])
            unless temp_is_same_type then
              is_same_type = false
            end
          end
        end
      end
    end
    is_same_type
  end

  #配列の中身の型チェック
  #配列の中身の型が異なればfalseを返す
  #空配列は中身がないのでtrueを返す
  def compare_array_type(array1, array2)
    is_same_type = true
    if array1.size == 0 || array2.size == 0 then
      return true
    end
    #TrueとFlaseは同じ型とする。
    array1_types = array1.map{|item|
      (item == true || item == false) ? true.class : item.class
    }.uniq
    array2_types = array2.map{|item|
      (item == true || item == false) ? true.class : item.class
    }.uniq
    if array1_types.size > 1 || array2_types.size > 1 then
      is_same_type = false #異なる型を入れるのは不可。どちらかに異なる型が２つ以上あればはfalseを返す
    else
      if array1.first == true || array1.first == false then
        is_same_type = false if array2.first != true && array2.first != false
      else
        if !array1.first.is_a?(array2.first.class) then
          is_same_type = false
        else
          if array1.first.is_a?(Hash) then
            is_same_type = compare_hash_type(array1.first, array2.first)
          elsif array1.first.is_a?(Array) then
            is_same_type = compare_array_type(array1.first, array2.first)
          end
        end
      end
    end
    is_same_type
  end

  #保存フィルタのための処理
  #バージョンのフィルタが異なる場合に、
  #現バージョンのフィルタが適用されるようにする
  def check_filters_type_by_type_source(type_source, filter)
    is_same_type = true
    if !type_source.is_a?(filter.class) then
      if type_source == true || type_source == false then
        is_same_type = false if filter != true && filter != false
      else
        is_same_type = false
      end
    else
      if type_source.is_a?(Hash) then
        is_same_type = compare_hash_type(type_source, filter)
      elsif type_source.is_a?(Array) then
        is_same_type = compare_array_type(type_source, filter)
      end
    end
    is_same_type
  end
  
  def establish_filters_consistency(filters, default_filters)
    filters = filters.deep_dup
    f_keys = default_filters.keys
    #無効なフィルタを削除
    (filters.keys - f_keys).each do |invalid_key|
      filters.delete(invalid_key)
    end
    #型が異なる/データが欠ける場合はデフォルト値にする
    f_keys.each do |f_key|
      default = default_filters[f_key][:default]
      if filters.has_key?(f_key) then
        if default_filters[f_key].has_key?(:type_source) then
          #デフォルト値に空配列がある場合や、ハッシュのvalueの型を指定したい場合、中身のデータ型が不明なため、ここで中身のデータをチェックする
          #その他の型/データ形式でも、設定可能な型をtype_sourceに指定して制限可能(type_sourceの値の型とデフォルト値の型が同じであれば、指定しても問題ないがここのチェックは無意味になる)
          is_type_pass = check_filters_type_by_type_source(default_filters[f_key][:type_source], filters[f_key])
          unless is_type_pass then
            filters[f_key] = default
            next
          end
        end
        if !default.is_a?(filters[f_key].class) then
          if default == true || default == false then
            filters[f_key] = default if filters[f_key] != true && filters[f_key] != false
          else
            filters[f_key] = default
          end
        else
          if default_filters[f_key][:values].is_a?(Array) then
            if !default_filters[f_key][:values].include?(filters[f_key]) then
              filters[f_key] = default
              next
            end
          end
          if default.is_a?(Hash) then
            is_same_type = compare_hash_type(default, filters[f_key])
            unless is_same_type then
              filters[f_key] = default
            end
          elsif default.is_a?(Array) then
            if default_filters[f_key][:values].is_a?(Hash) then
              filters[f_key] = filters[f_key].select{|f| default_filters[f_key][:values][f] || default_filters[f_key][:values][f.to_s]}
            end
            is_same_type = compare_array_type(default, filters[f_key])
            unless is_same_type then
              filters[f_key] = default
            end
          end
        end
      else
        filters[f_key] = default
      end
    end
    filters
  end

end
