# -*- coding: utf-8 -*-

require 'date'
require 'csv'
require 'tempfile'
require 'zip'

require 'sidfm_controller_support'
require 'uri'

class SidfmLocalVulsController < ApplicationController
  unloadable
  
  layout 'sidfm'

  self.main_menu = false

  protect_from_forgery  with: :exception
 
  helper :sidfm

  private

  include SidfmControllerSupport
  before_action :update_session_updated_at
  before_action :get_sidfm_roles
  before_action :create_common_data

  before_action :check_license

  before_action :check_roles
  before_action :check_is_user_admin, except: [:index, :export, :api_get_display_summary_rows, :api_get_display_summary_content]

  before_action :get_vm_version
  before_action :clear_import_session, except: [:index, :read_import_file, :progress_import, :import]
  before_action :clear_export_session, except: [:export]
  before_action -> { get_summary_table_lines('local_vuls') }

  def check_roles
    if !@sidfm_user || !@sidfm_roles.include?(:use) then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
  end

  def check_is_user_admin
    unless @sidfm_roles.include?(:admin) then
      redirect_to :action => :index, :notice => '許可されていない操作です。' and return
    end
  end

  def get_item_names
    item_names = {
      :vul_name => "脆弱性名",
      :area => "影響範囲",
      :risk => "危険度",
      :short_description => "簡易説明",
      :description => "概要",
      :flg_enable => "有効/無効",
      :product => "影響を受ける製品",
      :conf => "影響などの確認方法",
      :workaround => "対処方法 ",
      :related_info => "関連情報",
      :comment => "更新履歴",
      :cvss_score => "CVSSスコア",
      :cve_id => "CVE ID",
      :exploitable => "攻撃コードの有無",
      :cvss => "CVSS指標",
      :s_cat => "システムカテゴリ別のリスクの影響度(SRI)",
      :sri_memo => "SRI補足情報",
    }
  end

  def create_content
    content = []
    @all_vuls.each do |vul|
      #プロダクト
      products = {}
      lproducts = {}
      vul.sidfmd_local_vul_apps.each do |p|
        pid = p.pid
        vid = p.vid #vid == 0 はバージョン
        prod_info = SidfmdProduct.find_by(:pid => pid)
        if !prod_info then #プロダクトが削除された場合
          #p.delete
          next
        end
        if vid == 0 then
          fullname = ''
        else
          version_info = SidfmdVersion.find_by(:vid => vid)
          if !version_info then #バージョンが削除された場合
            #p.delete
            next
          else
            fullname = version_info.fullname
          end 
        end  
        products[pid] = [] if !products[pid]
        products[pid] << {
          :pid => pid.to_s,
          :name => prod_info.name,
          :vid => vid.to_s,
          :fullname => fullname
        }
      end
      vul.sidfmd_local_vul_lapps.each do |lp|
        pid = lp.sidfmd_lproduct_id
        vid = lp.sidfmd_lversion_id #vid == 0 はバージョン
        lprod_info = SidfmdLproduct.find_by(:id => pid)
        if !lprod_info || !lprod_info.flg_enable then #プロダクトが削除された場合 またはflg_enable == falseの場合
          #lp.delete
          next
        end
        if vid == 0 then
          fullname = ''
        else
          version_info =  SidfmdLversion.find_by(:id => vid)
          if !version_info then #バージョンが削除された場合
            #lp.delete
            next
          else
            fullname = version_info.fullname
          end
        end
        lproducts[pid] = [] if !lproducts[pid]
        lproducts[pid] << {
          :pid => pid.to_s,
          :name => lprod_info.name,
          :vid => vid.to_s,
          :fullname => fullname
        } 
      end
      #cve/cvss
      cve_ids = vul.sidfmd_local_vul_cves.order(:cve_id).map{|cve|
        {
          :cve_id => cve.cve_id,
          :cvss_score => cve.c3_base_score.to_s,
          :vectors => (
            (cve.c3_vector && cve.c3_vector.size > 0) ?
            cve.c3_vector.split('/').map{|v| v.split(':') }.to_h :
            CVSS_V3_VECTORS.keys.map{|v| [ v, nil ] }.to_h
          ),
          :exploitable => cve.c3_exploitable
        }
      }
      #sri
      sris = vul.sidfmd_local_vul_sris
      s_cats_sri = {}
      sris.each do |sri|
        s_cats_sri[sri.category_id.to_s] = sri.risk_id.to_s
      end
      content << {
        :id => vul.id,
        :html_id => vul.html_id,
        :vul_name => vul.title,
        :updated_at => vul.updated_at.strftime('%Y/%m/%d %H:%M'),
        :area_id => vul.area_id,
        :risk_id => vul.risk_id,
        :short_description => vul.short_description,
        :description => vul.description,
        :flg_enable => vul.flg_enable,
        :products => products,
        :lproducts => lproducts,
        :conf => vul.confirmation,
        :workaround => vul.workaround,
        :related_info => vul.related,
        :comments => vul.sidfmd_local_vul_logs.order(:id).map{|log| {:id => log.id, :mes => log.mes} },
        :cve_ids => cve_ids,
        :s_cats_sri => s_cats_sri,
        :sri_memo => (sris.first ? sris.first.memo : '')
      }
    end
    return content
  end

  #ソート
  def sort_vuls
    #order_param = @vuls[:order].split('_')
    last_index = @vuls[:order].rindex('_')
    col = @vuls[:order].slice(0..last_index - 1)
    order = @vuls[:order].slice(last_index + 1..@vuls[:order].size - 1)
    if col == 'name' then
      @all_vuls = @all_vuls.sort_by{|vul| vul.title.downcase}
      @all_vuls = @all_vuls.reverse if order == 'desc'
    elsif col == 'flg_enable' then
      if order == 'asc' then
        @all_vuls = @all_vuls.sort{|a, b|
          r = (a.flg_enable ? 0 : 1) <=> (b.flg_enable ? 0 : 1)
          if r == 0 then
            r = a.title.downcase <=> b.title.downcase
            if r == 0 then
              r = a.id <=> b.id
            end
          end
          r
        }
      else
        @all_vuls = @all_vuls.sort{|b, a|
          r = (a.flg_enable ? 0 : 1) <=> (b.flg_enable ? 0 : 1)
          if r == 0 then
            r = a.title.downcase <=> b.title.downcase
            if r == 0 then
              r = a.id <=> b.id
            end
          end
          r
        }
      end
    elsif col == 'updated_at' then
      if order == 'asc' then
        @all_vuls = @all_vuls.sort{|a, b|
          #秒切り捨て 
          r = a.updated_at.strftime('%Y-%m-%d %H:%M').to_time <=> b.updated_at.strftime('%Y-%m-%d %H:%M').to_time
          if r == 0 then
            r = a.title.downcase <=> b.title.downcase
            if r == 0 then
              r = a.id <=> b.id
            end
          end
          r 
        }
      else
        @all_vuls = @all_vuls.sort{|b, a|
          r = a.updated_at.strftime('%Y-%m-%d %H:%M').to_time <=> b.updated_at.strftime('%Y-%m-%d %H:%M').to_time
          if r == 0 then
            r = a.title.downcase <=> b.title.downcase
            if r == 0 then
              r = a.id <=> b.id
            end
          end
          r
        }
      end
    else #どれにも当てはまらない場合はvul.title ※普通の操作では起きない
      @all_vuls = @all_vuls.sort_by{|vul| vul.title.downcase}
    end
  end

  #検索
  def filter_vuls
    vul_name = @vuls[:search_vul_name]
    flg_enable = @vuls[:search_flg_enable]
    keyword = @vuls[:search_keyword]
    cve_id = @vuls[:search_cve_id]
    prod_info = @vuls[:search_product]
    is_vul_name = vul_name.size > 0
    is_flg_enable = flg_enable.size > 0
    is_keyword = keyword.size > 0
    is_cve_id = cve_id.size > 0
    is_product = prod_info.size > 0
    flg_enable = (flg_enable == 'enable' ? true : false)
    if is_vul_name || is_flg_enable || is_keyword || is_cve_id || is_product then
      filtered_vuls = []
      @all_vuls.each do |vul|
        if is_vul_name then
          vul_name = vul_name.downcase 
          next unless vul.title.downcase.index(vul_name)
        end
        if is_flg_enable then
          next unless vul.flg_enable == flg_enable
        end
        if is_cve_id then
          cve_id = cve_id.downcase
          cves = vul.sidfmd_local_vul_cves
          if cves.any? then
            next unless cves.select{|cve| cve.cve_id.downcase.index(cve_id)}.any?
          else
            next
          end
        end
        if is_keyword then
          keyword = keyword.downcase
          cves = vul.sidfmd_local_vul_cves
          cvss_score = (cves.any? ? cves.first.c3_base_score.to_s : nil) 
          sris = vul.sidfmd_local_vul_sris
          texts = [
            vul.title,
            vul.short_description,
            vul.description,
            vul.confirmation,
            vul.workaround,
            vul.related,
            vul.sidfmd_local_vul_sris.first.memo,
            vul.sidfmd_local_vul_logs.map{|log| log.mes}.join('')
          ]
          texts << cvss_score if cvss_score
          is_hit = false
          texts.each do |text|
            if text.downcase.index(keyword)
              is_hit = true
              break
            end
          end
          next unless is_hit
        end
        if is_product then
          is_hit_prod = false
          lprods = vul.sidfmd_local_vul_lapps
          if prod_info[:pid] then
            pid = prod_info[:pid].to_i
            #vul.sidfmd_local_vul_apps.where(:vid => 0).each do |p|
            vul.sidfmd_local_vul_apps.each do |p|
              if p.pid == pid then
                is_hit_prod = true
                break
              end 
            end
          elsif prod_info[:vid] then
            vid = prod_info[:vid].to_i
            vul.sidfmd_local_vul_apps.where.not(:vid => 0).each do |p|
              if p.vid == vid then
                is_hit_prod = true
                break
              end
            end
          elsif prod_info[:lpid] then
            lpid = prod_info[:lpid].to_i
            #vul.sidfmd_local_vul_lapps.where(:sidfmd_lversion_id => 0).each do |p|
            vul.sidfmd_local_vul_lapps.each do |p|
              if p.sidfmd_lproduct_id == lpid then
                is_hit_prod = true
                break 
               end
            end
          else
            lvid = prod_info[:lvid].to_i
            vul.sidfmd_local_vul_lapps.where.not(:sidfmd_lversion_id => 0).each do |p|
              if p.sidfmd_lversion_id == lvid then
                is_hit_prod = true
                break
              end 
            end 
          end
          next unless is_hit_prod 
        end
        filtered_vuls << vul
      end
      @all_vuls = filtered_vuls
    end
  end

  No_Strip_Items = [
    :title,
    :short_description,
    :description,
    :confirmation,
    :workaround,
    :related,
    :sri_memo,
    :log
  ]

  def strip_file_item_values(item)
    item.each do |key, val|
      if val && val.is_a?(String) && !No_Strip_Items.include?(key) then
        item[key] = val.strip
      end
    end
  end

  def create_data_from_files(files_hash)
    files = files_hash
    err = false
    #vul_info
    file_name = files['vul_info']['file_name']
    line_num = 2
    vul_info_lines = []
    vul_id_line = {}
    new_vul_id_line = {}
    area_names = SidfmdArea.pluck(:name)
    risk_names = SidfmdRisk.pluck(:name)
    sri_value_names = [
      SidfmdIssueSri::RISK_LOW ,
      SidfmdIssueSri::RISK_MEDIUM,
      SidfmdIssueSri::RISK_HIGH,
      SidfmdIssueSri::RISK_CRITICAL
    ]
    category_names = SidfmdCategory.where(:is_system => true).pluck(:name)

    CSV.parse(files['vul_info']['file'], :col_sep => "\t", :headers => :first_row).each do |line|
      item = {
        :new_vul_id => line['脆弱性登録ID'],
        :vul_id => line['脆弱性ID'],
        :title => line['脆弱性名'],
        :area => line['影響範囲'],
        :risk => line['危険度'],
        :short_description => line['簡易説明'],
        :description => line['概要'],
        :flg_enable => line['有効無効'],
        :confirmation => line['影響などの確認方法'],
        :workaround => line['対処方法'],
        :related => line['関連情報'],
        :category_risks => line['システムカテゴリ別のリスクの影響度'],
        :sri_memo => line['SRI補足情報'],
        :flg_delete => line['削除'],
        :line_num => line_num
      }
      strip_file_item_values(item)
      #入力チェック vul_infoファイルにエラーがある場合はimport確認画面に遷移させない
      #opt, registファイルにエラーがある場合はimport確認画面で警告表示
      if item[:new_vul_id] && item[:new_vul_id].size > 0 then
        if (item[:new_vul_id] =~ /^[0-9０-９]+$/) == nil then
          err = true
          session[:import_notices] << "#{file_name}の#{line_num}行目: 脆弱性登録IDは数字で入力して下さい。"
        else
          item[:new_vul_id] = item[:new_vul_id].to_i
          if new_vul_id_line[item[:new_vul_id]] then
            err = true
            session[:import_notices] << "#{file_name}の#{new_vul_id_line[item[:new_vul_id]]}行目と#{line_num}行目: 脆弱性登録IDが重複しています。"
          end
          new_vul_id_line[item[:new_vul_id]] = line_num
        end
      end

      if item[:vul_id] && item[:vul_id].size > 0 then
        if (item[:vul_id] =~ /^[0-9０-９]+$/) == nil then
          err = true
          session[:import_notices] << "#{file_name}の#{line_num}行目: 脆弱性IDは数字で入力して下さい。"
        end
        item[:vul_id] = item[:vul_id].to_i
        if vul_id_line[item[:vul_id]] then
          err = true
          session[:import_notices] << "#{file_name}の#{vul_id_line[item[:vul_id]]}行目と#{line_num}行目: 脆弱性IDが重複しています。"
        end

        if item[:flg_delete] && item[:flg_delete].size > 0 && item[:flg_delete].downcase != 'true' then
          err = true
          session[:import_notices] << "#{file_name}の#{line_num}行目: 削除は「TRUE」もしくは「true」を入力して下さい。"
        end

        item[:flg_delete] = item[:flg_delete].downcase if item[:flg_delete]
        if item[:flg_delete] == 'true'
          vul_info_lines << item
          line_num += 1
          next
        end

        if !item[:flg_enable] || item[:flg_enable].size == 0 then
          err = true
          session[:import_notices] << "#{file_name}の#{line_num}行目: 有効無効が入力されておりません。"
        end
        vul_id_line[item[:vul_id]] = line_num
      end

      if !item[:title] || item[:title].size == 0 then
        err = true
        session[:import_notices] << "#{file_name}の#{line_num}行目: 脆弱性名が入力されておりません。"
      end

      if !item[:area] || item[:area].size == 0 then
        err = true
        session[:import_notices] << "#{file_name}の#{line_num}行目: 影響範囲が入力されておりません。"
      elsif !area_names.include?(item[:area]) then
        err = true
        session[:import_notices] << "#{file_name}の#{line_num}行目: 影響範囲に誤りがあります。"
      end

      if !item[:risk] || item[:risk].size == 0 then
        err = true
        session[:import_notices] << "#{file_name}の#{line_num}行目: 危険度が入力されておりません。"
      elsif !risk_names.include?(item[:risk]) then
        err = true
        session[:import_notices] << "#{file_name}の#{line_num}行目: 危険度に誤りがあります。"
      end

      if item[:flg_enable] && item[:flg_enable].size > 0 && item[:flg_enable].downcase != 'false' && item[:flg_enable].downcase != 'true' then
        err = true
        session[:import_notices] << "#{file_name}の#{line_num}行目: 有効無効は「TRUE」「true」「FALSE」「false」のいずれかを入力して下さい。"
      end
      item[:flg_enable] = item[:flg_enable].downcase if item[:flg_enable]

      category_risks = item[:category_risks]
      if !category_risks then
        err = true
        session[:import_notices] << "#{file_name}の#{line_num}行目: システムカテゴリ別のリスクの影響度が入力されておりません。"
      else
        category_risks = category_risks.split(',')
        if category_risks.size != 8 then
          err = true
          session[:import_notices] << "#{file_name}の#{line_num}行目: システムカテゴリ別のリスクの影響度に誤りがあります。"
        else
          cat_err = false
          risk_err = false
          category_risks.each do |cat_risk|
            cat_risk = cat_risk.split(':')
            cat_name = cat_risk[0].strip
            risk_name = cat_risk[1].strip
            if !category_names.include?(cat_name) then
              cat_err = true
            elsif !sri_value_names.include?(risk_name) then
              risk_err = true
            end
          end
          if cat_err then
            err = true
            session[:import_notices] << "#{file_name}の#{line_num}行目: システムカテゴリ別のリスクの影響度のカテゴリ名に誤りがあります。"
          elsif risk_err then
            err = true
            session[:import_notices] << "#{file_name}の#{line_num}行目: システムカテゴリ別のリスクの影響度の影響度の値に誤りがあります。"
          end
        end
      end

      #リンクチェック
      if item[:description] && item[:description].size > 0 then
        link_converted_text = SidfmdLocalVul.get_link_converted_text(item[:description], '概要')
        if link_converted_text[:err] then
          err = true
          session[:import_notices] << "#{file_name}の#{line_num}行目: #{link_converted_text[:err].sub('「', '').sub('」', '')}"
        end
      end

      if item[:confirmation] && item[:confirmation].size > 0 then
        link_converted_text = SidfmdLocalVul.get_link_converted_text(item[:confirmation], '影響などの確認方法')
        if link_converted_text[:err] then      
          err = true
          session[:import_notices] << "#{file_name}の#{line_num}行目: #{link_converted_text[:err].sub('「', '').sub('」', '')}"
        end  
      end

      if item[:workaround] && item[:workaround].size > 0 then
        link_converted_text = SidfmdLocalVul.get_link_converted_text(item[:workaround], '対処方法')
        if link_converted_text[:err] then
          err = true
          session[:import_notices] << "#{file_name}の#{line_num}行目: #{link_converted_text[:err].sub('「', '').sub('」', '')}"
        end
      end

      if item[:related] && item[:related].size > 0 then
        link_converted_text = SidfmdLocalVul.get_link_converted_text(item[:related], '関連情報')
        if link_converted_text[:err] then
          err = true 
          session[:import_notices] << "#{file_name}の#{line_num}行目: #{link_converted_text[:err].sub('「', '').sub('」', '')}"
        end
      end

      vul_info_lines << item
      line_num += 1
    end
    if err then
      return {:err => err, :files => files}
    end
    files['vul_info'].delete('file')
    files['vul_info']['lines'] = vul_info_lines
    
    push_opt_item = ->(items_hash, item, file_name, line_num){
      #ファイルに誤りがある場合は、import確認画面で内容を表示(err = trueにしない)
      if !item[:vul_id] || item[:vul_id].size == 0 then
        session[:import_notices] << "#{file_name}の#{line_num}行目: 脆弱性IDが入力されておりません。"
      else
        if (item[:vul_id] =~ /^[0-9０-９]+$/) == nil then
          session[:import_notices] << "#{file_name}の#{line_num}行目: 脆弱性IDは数字で入力して下さい。"
        else
          item[:vul_id] = item[:vul_id].to_i
          items_hash[item[:vul_id]] = [] unless items_hash[item[:vul_id]]
          items_hash[item[:vul_id]] << item
        end
      end
      items_hash
    }

    if files['opt']['cve'].any? then
      #optフォルダ下のcveファイル
      cve_file = files['opt']['cve']['file']
      file_name = files['opt']['cve']['file_name']
      line_num = 2
      opt_cve_lines = {}
      CSV.parse(cve_file, :col_sep => "\t", :headers => :first_row).each do |line|
        item = {
          :vul_id => line['脆弱性ID'],
          :kind => line['種類'],
          :cvss_score => line['CVSSスコア'],
          :cve_id => line['CVEID'],
          :exploitable => line['攻撃コードの有無'],
          :vector => line['CVSS指標'],
          :line_num => line_num
        }
        strip_file_item_values(item)
        opt_cve_lines = push_opt_item.(opt_cve_lines, item, file_name, line_num)
        line_num += 1
      end
      files['opt']['cve'].delete('file')
      files['opt']['cve']['lines'] = opt_cve_lines
    end
    
    if files['opt']['product'].any? then
      #optフォルダ下のproductファイル
      prod_file = files['opt']['product']['file']
      file_name = files['opt']['product']['file_name']
      line_num = 2
      opt_prod_lines = {}
      CSV.parse(prod_file, :col_sep => "\t", :headers => :first_row).each do |line|
        item = {
          :vul_id => line['脆弱性ID'],
          :kind => line['種類'],
          :name => line['プロダクト名'],
          :line_num => line_num
        }
        strip_file_item_values(item)
        opt_prod_lines = push_opt_item.(opt_prod_lines, item, file_name, line_num)
        line_num += 1
      end
      files['opt']['product'].delete('file')
      files['opt']['product']['lines'] = opt_prod_lines
    end

    if files['opt']['change_log'].any? then
      #optフォルダ下のchangelogファイル
      log_file = files['opt']['change_log']['file']
      file_name = files['opt']['change_log']['file_name']
      line_num = 2
      opt_log_lines = {}
      CSV.parse(log_file, :col_sep => "\t", :headers => :first_row).each do |line|
        item = {
          :vul_id => line['脆弱性ID'],
          :kind => line['種類'],
          :log => line['更新履歴'],
          :order => line['順位'],
          :line_num => line_num
        }
        strip_file_item_values(item)
        opt_log_lines = push_opt_item.(opt_log_lines, item, file_name, line_num)
        line_num += 1
      end
      files['opt']['change_log'].delete('file')
      files['opt']['change_log']['lines'] = opt_log_lines
    end
    
    push_regist_item = ->(items_hash, item, file_name, line_num){
      #ファイルに誤りがある場合は、import確認画面で内容を表示(err = trueにしない)
      if !item[:new_vul_id] || item[:new_vul_id].size == 0 then
        session[:import_notices] << "#{file_name}の#{line_num}行目: 脆弱性IDが入力されておりません。"
      else
        if (item[:new_vul_id] =~ /^[0-9０-９]+$/) == nil then
          session[:import_notices] << "#{file_name}の#{line_num}行目: 脆弱性IDは数字で入力して下さい。"
        else
          item[:new_vul_id] = item[:new_vul_id].to_i
          items_hash[item[:new_vul_id]] = [] unless items_hash[item[:new_vul_id]]
          items_hash[item[:new_vul_id]] << item
        end
      end
      items_hash
    }
 
    if files['regist']['cve'].any? then
      #registフォルダ下のcveファイル
      cve_file = files['regist']['cve']['file']
      file_name = files['regist']['cve']['file_name']
      line_num = 2
      regist_cve_lines = {}
      CSV.parse(cve_file, :col_sep => "\t", :headers => :first_row).each do |line|
        item = {
          :new_vul_id => line['脆弱性登録ID'],
          :kind => line['種類'],
          :cvss_score => line['CVSSスコア'],
          :cve_id => line['CVEID'],
          :exploitable => line['攻撃コードの有無'],
          :vector => line['CVSS指標'],
          :line_num => line_num
        }
        strip_file_item_values(item)
        regist_cve_lines = push_regist_item.(regist_cve_lines, item, file_name, line_num)
        line_num += 1
      end
      files['regist']['cve'].delete('file')
      files['regist']['cve']['lines'] = regist_cve_lines
    end

    if files['regist']['product'].any? then
      #registフォルダ下のproductファイル
      prod_file = files['regist']['product']['file']
      file_name = files['regist']['product']['file_name']
      line_num = 2
      regist_prod_lines = {}
      CSV.parse(prod_file, :col_sep => "\t", :headers => :first_row).each do |line|
        item = {
          :new_vul_id => line['脆弱性登録ID'],
          :kind => line['種類'],
          :name => line['プロダクト名'],
          :line_num => line_num
        }
        strip_file_item_values(item)
        regist_prod_lines = push_regist_item.(regist_prod_lines, item, file_name, line_num)
        line_num += 1
      end
      files['regist']['product'].delete('file')
      files['regist']['product']['lines'] = regist_prod_lines
    end

    if files['regist']['change_log'].any? then
      #registフォルダ下のchangelogファイル
      log_file = files['regist']['change_log']['file']
      file_name = files['regist']['change_log']['file_name']
      line_num = 2
      regist_log_lines = {}
      CSV.parse(log_file, :col_sep => "\t", :headers => :first_row).each do |line|
        item = {
          :new_vul_id => line['脆弱性登録ID'],
          :kind => line['種類'],
          :log => line['更新履歴'],
          :order => line['順位'],
          :line_num => line_num
        }
        strip_file_item_values(item)
        regist_log_lines = push_regist_item.(regist_log_lines, item, file_name, line_num)
        line_num += 1
      end
      files['regist']['change_log'].delete('file')
      files['regist']['change_log']['lines'] = regist_log_lines
    end
    
    return { :err => err, :files => files }
  end

  def filter_cves(cves, file_name)
    is_pass = true
    filtered_cves = []
    cves.each do |cve|
      cve = cve.symbolize_keys
      if !cve[:kind] || cve[:kind].size == 0 then
        is_pass = false
        session[:import_notices] << "#{file_name}の#{cve[:line_num]}行目: 種類が入力されておりません。"
      elsif cve[:kind] != 'cve' then
        is_pass = false
        session[:import_notices] << "#{file_name}の#{cve[:line_num]}行目: 種類に誤りがあります。"
      elsif cve[:cvss_score] and cve[:cvss_score].size > 0 and ((cve[:cvss_score] =~ /^[0-9０-９]$|^[0-9０-９]+\.[0-9０-９]?$|10|１０/) == nil) then
        is_pass = false
        session[:import_notices] << "#{file_name}の#{cve[:line_num]}行目: CVSSスコアは0.0以上10.0以下の数字で、小数点以下を指定する場合は第一位までを入力して下さい。"
      elsif cve[:cvss_score] and cve[:cvss_score].size > 0 and (cve[:cvss_score].to_f > 10.0 or cve[:cvss_score].to_f < 0.0) then
        is_pass = false
        session[:import_notices] << "#{file_name}の#{cve[:line_num]}行目: CVSSスコアは0.0以上10.0以下の数字で入力して下さい。"
      elsif !cve[:exploitable] || (cve[:exploitable].downcase != 'true' && cve[:exploitable].downcase != 'false') then
        is_pass = false
        session[:import_notices] << "#{file_name}の#{cve[:line_num]}行目: 攻撃コードは「TRUE」「true」「FALSE」「false」のいずれかを入力して下さい。"
      elsif !cve[:cve_id] || cve[:cve_id].size == 0 then
        is_pass = false
        session[:import_notices] << "#{file_name}の#{cve[:line_num]}行目: CVEIDが入力されておりません。"
      elsif (cve[:vector] && cve[:vector].size > 0) && (!cve[:cvss_score] || cve[:cvss_score].size == 0) then
        is_pass = false
        session[:import_notices] << "#{file_name}の#{cve[:line_num]}行目: CVSS指標を登録する場合は、CVSSスコアの入力が必須です。"
      elsif (cve[:cvss_score] && cve[:cvss_score].size > 0) && (!cve[:vector] || cve[:vector].size == 0) then
        is_pass = false
        session[:import_notices] << "#{file_name}の#{cve[:line_num]}行目: CVSSスコアを登録する場合は、CVSS指標の入力が必須です。"   
      else
        if cve[:vector] && cve[:vector].size > 0 then 
          cve[:vector] = cve[:vector].upcase.strip
          is_vector_correct = check_cve_vector(cve[:vector])
          if !is_vector_correct then
            is_pass = false
            session[:import_notices] << "#{file_name}の#{cve[:line_num]}行目: CVSS指標に誤りがあります。"
          end
        end
        cve[:vector] = nil if cve[:vector] and cve[:vector].size == 0 
        cve[:cvss_score] = nil if cve[:cvss_score] and cve[:cvss_score].size == 0
      
        if is_pass then
          cve[:exploitable] = cve[:exploitable].downcase
          cve[:cve_id] = cve[:cve_id].tr('０-９', '0-9').strip.upcase
          if(cve[:cve_id] =~ /^CVE-[0-9]{4}-[0-9]+$/) == nil then
            is_pass = false
            session[:import_notices] << "#{file_name}の#{cve[:line_num]}行目: CVEIDに誤りがあります。"
          end
          if cve[:cvss_score] && cve[:cvss_score].size > 0 then
            cve[:cvss_score] = cve[:cvss_score].tr('０-９', '0-9')
            if cve[:cvss_score].index('10') && cve[:cvss_score] != '10' && cve[:cvss_score] != '10.0' then
              is_pass = false
              session[:import_notices] << "#{file_name}の#{cve[:line_num]}行目: CVSSスコアは0.0以上10.0以下の数字で、小数点以下を指定する場合は第一位までを入力して下さい。"
            else
              cve[:cvss_score] = cve[:cvss_score].to_f
            end
          end
          same_cve_id = filtered_cves.find{|f_cve| f_cve[:cve_id] == cve[:cve_id] }
          if same_cve_id then
            is_pass = false
            session[:import_notices] << "#{file_name}の#{same_cve_id[:line_num]}行目と#{cve[:line_num]}行目: CVE IDが重複しています。"
          end
          filtered_cves << cve if is_pass
        end
      end
    end
    filtered_cves = filtered_cves.sort_by{|cve_id| cve_id[:cve_id]}
    { :is_pass => is_pass, :cves => filtered_cves }
  end

  def filter_prods(prods, file_name)
    filtered_prods = []
    prods.each do |prod|
      prod = prod.symbolize_keys
      if !prod[:kind] || prod[:kind].size == 0 then
        session[:import_notices] << "#{file_name}の#{prod[:line_num]}行目: 種類が入力されておりません。"
        next
      elsif prod[:kind] != 'manual_product' && prod[:kind] != 'local_product' then
        session[:import_notices] << "#{file_name}の#{prod[:line_num]}行目: 種類に誤りがあります。"
        next
      elsif !prod[:name] || prod[:name].strip.size == 0 then
        session[:import_notices] << "#{file_name}の#{prod[:line_num]}行目: プロダクト名が入力されておりません。"
        next
      end
      prod[:name] = prod[:name].strip
      if prod[:kind] == 'manual_product' then
        ver = SidfmdVersion.find_by(:fullname => prod[:name]) 
        if ver then
          filtered_prods << prod
        else
          prd = SidfmdProduct.find_by(:name => prod[:name])
          if prd then
            filtered_prods << prod
          else
            session[:import_notices] << "#{file_name}の#{prod[:line_num]}行目: プロダクト名に誤りがあるか、データベースに存在しないプロダクト名です。"
          end
        end
      else
        lver = SidfmdLversion.find_by(:fullname => prod[:name])
        if lver then
          filtered_prods << prod
        else
          lprd = SidfmdLproduct.find_by(:name => prod[:name])
          if lprd then
            filtered_prods << prod
          else
            session[:import_notices] << "#{file_name}の#{prod[:line_num]}行目: プロダクト名に誤りがあるか、データベースに存在しないプロダクト名です。"
          end
        end
      end
    end
    filtered_prods = filtered_prods.uniq{|prd| prd[:name]}
  end

  def filter_logs(logs, file_name, is_regist)
    filtered_logs = []
    log_order_line = {}
    logs.each do |log|
      log = log.symbolize_keys
      if !log[:kind] || log[:kind].size == 0 then
        session[:import_notices] << "#{file_name}の#{log[:line_num]}行目: 種類が入力されておりません。"
        next
      elsif log[:kind] != 'change_log' then
        session[:import_notices] << "#{file_name}の#{log[:line_num]}行目: 種類に誤りがあります。"
        next
      elsif !log[:log] || log[:log].size == 0 then
        session[:import_notices] << "#{file_name}の#{log[:line_num]}行目: 更新履歴が入力されておりません。"
        next
      elsif !log[:order] || log[:order].size == 0 then
        session[:import_notices] << "#{file_name}の#{log[:line_num]}行目: 順位が入力されておりません。"
        next
      elsif (log[:order] =~ /^[0-9０-９]+$/) == nil then
        session[:import_notices] << "#{file_name}の#{log[:line_num]}行目: 順位は数字で入力して下さい。"
        next
      end
      log[:order] = log[:order].tr('０-９', '0-9').to_i
      if log_order_line[log[:order]] then
        session[:import_notices] << "#{file_name}の#{log_order_line[log[:order]]}行目と#{log[:line_num]}行目: 順位が重複しています。"
        next
      end
      log_order_line[log[:order]] = log[:line_num]
      filtered_logs << log
    end

    filtered_logs = filtered_logs.sort_by{|log| log[:order]}

    if is_regist then
      order = 1
      #順位の不整合が無いことを保証する処理 #変更の場合はその順位のものをアップデートするため、この処理は行わない
      filtered_logs.each do |log|
        log[:order] = order
        order += 1
      end
    end

    filtered_logs 
  end

  def create_vul_info_option(old_value, new_value, action, action_name)
    option = {
      :action => action,
      :new_data => new_value,
      :old_data => old_value,
      :action_name => action_name
    }
  end

  def create_cve_option(current_cves, new_cves)
    #first_current_cve = current_cves.first
    #if item_name == 'CVSSスコア' then
    #  old_cve = first_current_cve ? first_current_cve.c3_base_score : nil
    #  new_cve = new_cve ? new_cve[:cvss_score] : nil
    #  action_name = 'cvss_score_change'
    #elsif item_name == 'CVE ID' then
    #  old_cve = first_current_cve ? current_cves.map{|cve| cve.cve_id}.sort : nil
    #  new_cve = (new_cve and new_cve[:cve_ids]) ? new_cve[:cve_ids].split(',') : nil
    #  action_name = 'cve_ids_change'
    #elsif item_name == '攻撃コードの有無' then
    #  old_cve = (first_current_cve ? (first_current_cve.c3_exploitable ? first_current_cve.c3_exploitable.to_s : nil) : nil)
    #  new_cve = new_cve ? new_cve[:exploitable] : nil
    #  action_name = 'exploitable_change'
    #elsif item_name == 'CVSS指標' then
    #  old_cve = (first_current_cve ? (first_current_cve.c3_vector ? first_current_cve.c3_vector.split('/').map{|v| v.split(':')}.to_h : nil) : nil)
    #  new_cve = (new_cve ? (new_cve[:vector] ? new_cve[:vector].split('/').map{|v| v.split(':')}.to_h : nil) : nil)
    #  action_name = 'vector_change'
    #end
    option = {
      :action => 'cve_ids_change',
      :new_data => new_cves,
      :old_data => current_cves.values.sort_by{|cve| cve[:cve_id]},
      :action_name => 'CVE ID/CVSS情報変更'
    }
  end

  def create_product_option(old_app_prds, old_app_vers, old_lapp_prds, old_lapp_vers, new_products)
    old_prds = []
    old_app_prds.each do |prd|
      old_prds << { :name => prd.name, :kind => 'manual_product' }
    end
    old_app_vers.each do |ver|
      old_prds << { :name => ver.fullname, :kind => 'manual_product' }
    end

    old_lprds = []
    old_lapp_prds.each do |lprd|
      old_lprds << { :name => lprd.name, :kind => 'local_product' }
    end
    old_lapp_vers.each do |lver|
      old_lprds << { :name => lver.fullname, :kind => 'local_product' }
    end

    old_prds = old_prds.uniq{|prd| prd[:name]}.sort_by{|prd| prd[:name]}
    old_lprds = old_lprds.uniq{|prd| prd[:name]}.sort_by{|lprd| lprd[:name]}
    if new_products then 
      new_products = new_products.sort{|a, b|
        r = (a[:kind] == 'manual_product' ? 100 : 1000) <=> (b[:kind] == 'manual_product' ? 100 : 1000)
        if r == 0 then
          r = a[:name] <=> b[:name]
        end
        r
      }
    end
    option = {
      :action => 'product_change',
      :new_data => new_products,
      :old_data => old_prds + old_lprds,
      :action_name => '影響を受ける製品変更'
    }
  end

  def create_log_option(current_logs, new_logs)
    option = {
      :action => 'logs_change',
      :new_data => new_logs,
      :old_data => current_logs,
      :action_name => '更新履歴変更'
    }
  end 

  def check_cve_vector(vector)
    cve_vector_values = {
      'AV' => [ 'P', 'L', 'A', 'N'],
      'AC' => ['H', 'L'],
      'PR' => ['H', 'L', 'N'],
      'UI' => ['R', 'N'],
      'S' => ['U', 'C'],
      'C' => ['N', 'L', 'H'],
      'I' => ['N', 'L', 'H'],
      'A' => ['N', 'L', 'H']
    }
    vectors = vector.split('/').map{|v| v.split(':')}
    is_vector_correct = true
    cnt_vectors = 0
    vector_keys = []
    vectors.each do |v|
      if v.size != 2 then
        is_vector_correct = false
        break
      end
      values = cve_vector_values[v[0]]
      if !values || !values.include?(v[1]) || vector_keys.include?(v[0]) then
        is_vector_correct = false
        break
      else
        cnt_vectors += 1
        vector_keys << v[0]
      end
    end
    is_vector_correct = (cnt_vectors == 8) if is_vector_correct

    is_vector_correct
  end

  #import_applyでcve変更
=begin
  def update_cve_by_options(vul_id, new_data, action_name, options, is_cve_created, is_cve_destroyed)
    r = {}
    cves = SidfmdLocalVulCve.where(:sidfmd_local_vul_id => vul_id)
    if cves.empty? && !is_cve_created && !is_cve_destroyed then
      #未登録から登録
      option_data = {}
      options.each do |option|
        item_name = option[:action].sub('_change', '')
        option_data[item_name] = option[:new_data]
      end
      #問題なければ登録
      if option_data['cve_ids'] && option_data['exploitable'] && 
        ((option_data['cvss_score'] && option_data['vector']) || (!option_data['cvss_score'] && !option_data['vector']))
      then
        cvss_score = option_data['cvss_score'] ? option_data['cvss_score'].to_f : nil
        if cvss_score then
          if cvss_score > 10 || cvss_score < 0 then
            logger.info('invalid cvss_score')
            r[:err] = true
          end
        end
        vector = option_data['vector'] ? option_data['vector'].map{|k, v| "#{k.to_s}:#{v}"}.join('/') : nil
        if vector then
          is_vector_correct = check_cve_vector(vector)
          if !is_vector_correct then
            logger.info('invalid cve_vector')
            r[:err] = true
          end
        end
        exploitable = (option_data['exploitable'].to_s == 'true' ? true : false)
        new_cves = []
        is_cve_id_correct = true
        option_data['cve_ids'].each do |cve_id|
          if(cve_id =~ /^CVE-[0-9]{4}-[0-9]+$/) == nil then
            is_cve_id_correct = false
            r[:err] = true
            break
          end
          new_cves << SidfmdLocalVulCve.new(
            sidfmd_local_vul_id: vul_id,
            cve_id: cve_id,
            c3_base_score: cvss_score,
            c3_vector: vector,
            c3_exploitable: exploitable
          )
        end
        if is_cve_id_correct then
          SidfmdLocalVulCve.import new_cves
          r[:is_cve_created] = true
        else
          r[:err] = true
        end
      end
    elsif action_name == 'cve_ids_change' && cves.any? && !new_data then
      #cveの登録を削除
      cves.destroy_all
      r[:is_cve_destroyed] = true
    elsif cves && !is_cve_created then
      #変更
      case action_name
      when 'cvss_score_change' then
        cvss_score = new_data ? new_data.to_f : nil
        if cvss_score && (cvss_score > 10 || cvss_score < 0) then
          logger.info('invalid cvss_score')
          r[:err] = true
        else
          cves.update_all(:c3_base_score => cvss_score)
        end
      when 'cve_ids_change' then
        first_cve = cves.first
        vul_id = first_cve.sidfmd_local_vul_id
        cvss_score = first_cve.c3_base_score
        vector = first_cve.c3_vector
        exploitable = first_cve.c3_exploitable

        cves.where.not(:cve_id => new_data).destroy_all
        current_cve_ids = cves.map{|cve| cve.cve_id}
        new_cves = []
        new_data.each do |cve_id|
          if(cve_id =~ /^CVE-[0-9]{4}-[0-9]+$/) == nil then
            is_cve_id_correct = false
            r[:err] = true
            break
          end
          if !current_cve_ids.include?(cve_id) then
            new_cves << SidfmdLocalVulCve.new(
              sidfmd_local_vul_id: vul_id,
              cve_id: cve_id,
              c3_base_score: cvss_score,
              c3_vector: vector,
              c3_exploitable: exploitable
            )
          end
        end
 
        SidfmdLocalVulCve.import(new_cves) unless r[:err]
      when 'vector_change' then
        vector = new_data ? new_data.map{|k, v| "#{k.to_s}:#{v}"}.join('/') : nil
        is_vector_correct = ((vector == nil) or check_cve_vector(vector))
        if !is_vector_correct then
          logger.info('invalid cve_vector')
          r[:err] = true
        else
          cves.update_all(:c3_vector => vector)
        end
      when 'exploitable_change' then
        cves.update_all(:c3_exploitable => new_data.to_s == 'true' ? true : false)
      end
    end

    r
  end
=end

  FILTER_KEYS = {}

  def init_filter_keys
    if FILTER_KEYS.size == 0 then
      FILTER_KEYS[:search_vul_name] = ''
      FILTER_KEYS[:search_flg_enable] = ''
      FILTER_KEYS[:search_keyword] = ''
      FILTER_KEYS[:search_cve_id] = ''
      FILTER_KEYS[:product] = {}
    end
  end

  def check_filters
    init_filter_keys
    FILTER_KEYS.each do |key, val|
      unless @vuls[:filters].has_key?(key) then
        @vuls[:filters][key] = val
      end
    end
    @default_filters = FILTER_KEYS
  end

  def parse_params
    f_params = params.fetch(:pages, {}).permit(:offset, :per_page, :order, :filters, :checked, :all_checked)
    @vuls = {}
    @vuls[:per_page] = (@summary_table_lines || f_params[:per_page] || 10).to_i
    @vuls[:offset] = (f_params[:offset] || 0).to_i
    @vuls[:filters] = JSON.parse(f_params[:filters] || '{}').symbolize_keys
    @vuls[:checked] = JSON.parse(f_params[:checked] || '[]')
    @vuls[:all_checked] = ((f_params[:all_checked] || '') == 'true')
    @vuls[:order] = (f_params[:order] || 'name_asc')
    @vuls[:search_vul_name] = (@vuls[:filters][:search_vul_name] || '')
    @vuls[:search_flg_enable] = (@vuls[:filters][:search_flg_enable] || '')
    @vuls[:search_keyword] = (@vuls[:filters][:search_keyword] || '')
    @vuls[:search_cve_id] = (@vuls[:filters][:search_cve_id] || '')
    @vuls[:search_product] = (@vuls[:filters][:product] || {}).symbolize_keys
  end

  def create_summary_lvul_html(vul, **opts)
    data = {
      :chkbox => view_context.create_summary_local_vul_chkbox(vul[:id], opts[:checked]),
      :name => view_context.create_summary_local_vul_name(vul[:vul_name]),
      :updated_at => view_context.create_summary_local_vul_updated_at(vul[:updated_at]),
      :flg_enable => view_context.create_summary_local_vul_status(vul[:flg_enable]),
      :detail => view_context.create_summary_local_vul_detail(vul[:html_id]),
      :edit => view_context.create_summary_local_vul_edit(vul, @sidfm_roles)
    }
  end

  CVSS_V3_VECTORS = SidfmdIssueCve.get_cvss_v3_vectors

  def create_regist_cvss_vectors(vectors_hash)
    vectors = []
    if vectors_hash then
      vectors_hash.each do |key, value|
        if !CVSS_V3_VECTORS[key.to_s] or !CVSS_V3_VECTORS[key.to_s]['values'][value.to_s] then
          vectors = [] #validation
          break
        end
        vectors << "#{key}:#{value}"
      end
    end
    vectors
  end

  public 

  def index
    f_params = params.fetch(:pages, {}).permit(:clear_session)
    if f_params[:clear_session] then
      clear_import_session
      redirect_to :action => :index
    end
    parse_params
    check_filters

    @areas = SidfmdArea.order('id')
    @risks = SidfmdRisk.order('risk_id')
    @sris = {
      #SidfmdIssueSri::VALUE_NONE => SidfmdIssueSri::RISK_NONE,
      SidfmdIssueSri::VALUE_LOW => SidfmdIssueSri::RISK_LOW ,
      SidfmdIssueSri::VALUE_MEDIUM => SidfmdIssueSri::RISK_MEDIUM,
      SidfmdIssueSri::VALUE_HIGH => SidfmdIssueSri::RISK_HIGH,
      SidfmdIssueSri::VALUE_CRITICAL => SidfmdIssueSri::RISK_CRITICAL
    }
    @s_cats_data = SidfmdCategory.system
    #SidfmdCategory.where(:is_system => true).order('id').each do |s_cat|
    #  @s_cats_data << { :id => s_cat.id, :name => s_cat.name }
    #end
    @item_names = get_item_names
    @enable_flgs = {
      '有効' => 'enable',
      '無効' => 'disable'
    }
    
    @import_notices = session[:import_notices]
    clear_import_session

  end

  def create
    f_params = params.permit(
      :vul_name,
      :area_id,
      :risk_id,
      :short_description,
      :description,
      :flg_enable,
      :affected_products,
      :affected_lproducts,
      :confirmation,
      :workaround,
      :related_info,
      :cve_ids,
      :s_categories_sri,
      :sri_memo
    )
    vul_name = f_params[:vul_name]
    area_id = f_params[:area_id]
    risk_id = f_params[:risk_id]
    short_description = f_params[:short_description]
    description = f_params[:description]
    flg_enable = (f_params[:flg_enable] == 'enable' ? true : false)
    begin 
      affected_products = JSON.parse(f_params[:affected_products])
      affected_lproducts = JSON.parse(f_params[:affected_lproducts])
      cve_ids = JSON.parse(f_params[:cve_ids], :symbolize_names => true)
      s_cats_sri = JSON.parse(f_params[:s_categories_sri])
    rescue JSON::ParserError => e
      logger.info(e)
      render :layout => false, :json => { :err_message => '不正な操作です' } and return
    end
    confirmation = f_params[:confirmation]
    workaround  = f_params[:workaround]
    related_info = f_params[:related_info]
    flg_exploitable = (f_params[:flg_exploitable].to_s == 'true' ? true : false) 
    sri_memo = f_params[:sri_memo]
    if !vul_name || !area_id || !risk_id || !description || !short_description || flg_enable == nil || !confirmation || !workaround || !related_info || flg_exploitable == nil || !sri_memo
    then
      #logger.info([vul_name , area_id, risk_id, short_description, description, flg_enable, confirmation_text, workaround_text, related_info, comment, sri_memo].inspect)
      render :layout => false, :json => { :err_message => '不正な操作です' } and return
    end
    begin
      SidfmdLocalVul.transaction do 
        new_vul = SidfmdLocalVul.new(
                    flg_enable: flg_enable,
                    title: vul_name,
                    risk_id: risk_id,
                    area_id: area_id,
                    short_description: short_description,
                    description: description,
                    confirmation: confirmation,
                    workaround: workaround,
                    related: related_info,
                  )
        if new_vul.save then
          #更新履歴
          time_now = Time.now
          SidfmdLocalVulLog.create(
            sidfmd_local_vul_id: new_vul.id, 
            dt: time_now,
            mes: "#{time_now.strftime('%Y/%m/%d')} コンテンツ 登録"
          )
          #プロダクト
          affected_products.each do |prod_id, data_array|
            data_array.each do |data|
              SidfmdLocalVulApp.create(
                sidfmd_local_vul_id: new_vul.id,
                pid: prod_id.to_i,
                vid: data['vid'].to_i #バージョンの場合はvid > 0で、プロダクトの場合はvid == 0
              )
            end
          end
          #ローカルプロダクト
          affected_lproducts.each do |lprod_id, data_array|
            data_array.each do |data|
              SidfmdLocalVulLapp.create(
                sidfmd_local_vul_id: new_vul.id,
                sidfmd_lproduct_id: lprod_id.to_i,
                sidfmd_lversion_id: data['vid'].to_i #バージョンの場合はvid > 0で、プロダクトの場合はvid == 0
              )
            end
          end
          #CVE ID
          cve_ids.uniq{|cve_id| cve_id[:cve_id]}.each do |cve_id|
            unless cve_id[:cve_id] =~ /\ACVE\-[0-9]{4}\-[0-9]+\z/ then
              raise Exception #不正
            end
            regist_vectors = create_regist_cvss_vectors cve_id[:vectors]
            SidfmdLocalVulCve.create(
              sidfmd_local_vul_id: new_vul.id,
              cve_id: cve_id[:cve_id],
              c3_base_score: (regist_vectors.any? ? sprintf('%2.1f', cve_id[:cvss_score].to_f) : nil),
              c3_vector: ((regist_vectors.any? and cve_id[:cvss_score]) ? regist_vectors.join('/') : nil),
              c3_exploitable: cve_id[:exploitable].to_s == 'true'
            )  
          end
          #SRI
          s_cats_sri.each do |cat_id, risk_id|
            SidfmdLocalVulSri.create(
              sidfmd_local_vul_id: new_vul.id,
              category_id: cat_id,
              risk_id: risk_id,
              memo: sri_memo
            )
          end
        end
      end
      Rails.cache.clear
    rescue Exception => e 
      logger.info(e)
      render :layout => false, :json => { :err_message => '脆弱性の登録に失敗しました。' } and return
    end
    render :layout => false, :json => { :err_message => '' }
  end

  def modify
    f_params = params.permit(
      :id,
      :vul_name,
      :area_id,
      :risk_id,
      :short_description,
      :description,
      :flg_enable,
      :affected_products,
      :affected_lproducts,
      :confirmation,
      :workaround,
      :related_info,
      :cve_ids,
      :s_categories_sri,
      :sri_memo,
      :comments
    )
    vul_id = f_params[:id]
    vul_name = f_params[:vul_name]
    area_id = f_params[:area_id]
    risk_id = f_params[:risk_id]
    short_description = f_params[:short_description]
    description = f_params[:description]
    flg_enable = (f_params[:flg_enable] == 'enable' ? true : false)
    begin
      affected_products = JSON.parse(f_params[:affected_products])
      affected_lproducts = JSON.parse(f_params[:affected_lproducts])
      cve_ids = JSON.parse(f_params[:cve_ids], :symbolize_names => true)
      s_cats_sri = JSON.parse(f_params[:s_categories_sri])
      comments = JSON.parse(f_params[:comments])
    rescue JSON::ParserError => e
      logger.info(e)
      render :layout => false, :json => { :err_message => '不正な操作です' } and return
    end
    confirmation = f_params[:confirmation]
    workaround  = f_params[:workaround]
    related_info = f_params[:related_info]
    flg_exploitable = (f_params[:flg_exploitable].to_s == 'true' ? true : false)
    sri_memo = f_params[:sri_memo]
    if !vul_id || !vul_name || !area_id || !risk_id || !description || !short_description || flg_enable == nil || !confirmation || !workaround || !related_info || flg_exploitable == nil || !sri_memo
    then
      render :layout => false, :json => { :err_message => '不正な操作です' } and return
    end
    
    begin
      SidfmdLocalVul.transaction do
        time_now = Time.now
        mod_vul = SidfmdLocalVul.find(vul_id)
        mod_vul.update_attributes(
          flg_enable: flg_enable,
          title: vul_name,
          risk_id: risk_id,
          area_id: area_id,
          short_description: short_description,
          description: description,
          confirmation: confirmation,
          workaround: workaround,
          related: related_info,
          updated_at: time_now
        )
        #更新履歴
        vul_comments = mod_vul.sidfmd_local_vul_logs
        comments.each do |cmt|
          cmt = cmt.symbolize_keys
          target_cmt = vul_comments.find_by(:id => cmt[:id])
          if target_cmt then
            if target_cmt.mes != cmt[:mes] then
              target_cmt.update_attributes(:dt => time_now, :mes => cmt[:mes])
            end
          else#id == 0は新しい更新履歴
            SidfmdLocalVulLog.create(
              sidfmd_local_vul_id: mod_vul.id,
              dt: time_now,
              mes: cmt[:mes]
            )
          end
        end
        #プロダクト
        mod_vul.sidfmd_local_vul_apps.destroy_all
        affected_products.each do |prod_id, data_array|
          data_array.each do |data|
            SidfmdLocalVulApp.create(
              sidfmd_local_vul_id: mod_vul.id,
              pid: prod_id.to_i,
              vid: data['vid'].to_i #バージョンの場合はvid > 0で、プロダクトの場合はvid == 0
            )
          end
        end
        #ローカルプロダクト
        mod_vul.sidfmd_local_vul_lapps.destroy_all
        affected_lproducts.each do |lprod_id, data_array|
          data_array.each do |data|
            SidfmdLocalVulLapp.create(
              sidfmd_local_vul_id: mod_vul.id,
              sidfmd_lproduct_id: lprod_id.to_i,
              sidfmd_lversion_id: data['vid'].to_i #バージョンの場合はvid > 0で、プロダクトの場合はvid == 0
            )
          end
        end
        #CVE ID
        mod_vul.sidfmd_local_vul_cves.destroy_all
        cve_ids.uniq{|cve_id| cve_id[:cve_id]}.each do |cve_id|
          unless cve_id[:cve_id] =~ /\ACVE\-[0-9]{4}\-[0-9]+\z/ then
            raise Exception #不正
          end
          regist_vectors = create_regist_cvss_vectors cve_id[:vectors]
          SidfmdLocalVulCve.create(
            sidfmd_local_vul_id: mod_vul.id,
            cve_id: cve_id[:cve_id],
            c3_base_score: (regist_vectors.any? ? sprintf('%2.1f', cve_id[:cvss_score].to_f) : nil),
            c3_vector: ((regist_vectors.any? and cve_id[:cvss_score]) ? regist_vectors.join('/') : nil),
            c3_exploitable: cve_id[:exploitable].to_s == 'true'
          )
        end
        #SRI
        mod_vul.sidfmd_local_vul_sris.destroy_all
        s_cats_sri.each do |cat_id, risk_id|
          SidfmdLocalVulSri.create(
            sidfmd_local_vul_id: mod_vul.id,
            category_id: cat_id,
            risk_id: risk_id,
            memo: sri_memo
          )
        end
      end
    rescue Exception => e
      logger.info(e)
      render :layout => false, :json => { :err_message => '脆弱性の変更に失敗しました。' } and return
    end
    Rails.cache.clear
    render :layout => false, :json => { :err_message => '' }
  end

  def delete
    #f_params = params.fetch(:form, {}).permit(:delete_id)
    f_params = params.permit(:delete_id)
    delete_id = f_params[:delete_id]
    err_message = ''
    host_ids = []
    html_id = 0
    SidfmdLocalVul.transaction do 
      begin 
        vul = SidfmdLocalVul.find(delete_id)
        html_id = vul.html_id
        #view関連のホストidを取得
        host_ids = SidfmdVulView.relate_hostids([ html_id ]) 
        vul.destroy
        issue_ids = SidfmdIssue.where(:html_id => html_id).pluck(:issue_id)
        if issue_ids.any? then
          SidfmdIssueCtag.where(:sidfmd_issue_id => issue_ids).delete_all
          Issue.where(:id => issue_ids).destroy_all
        end
      rescue Exception => e 
        logger.info(e)
        err_message = '脆弱性の削除に失敗しました。'
        render :layout => false, :json => { :err_message => err_message } and return
      end
    end
    r = {
      :err_message => err_message,
      :host_ids => host_ids,
      :html_id => html_id  
    }
    Rails.cache.clear
    render :layout => false, :json => r
  end

  #一括有効無効
  def bulk_change_status
    f_params = params.permit(:ids, :flg_enable) # :offset, :per_page, :order, :filters)
    begin 
      ids = JSON.parse(f_params[:ids])
    rescue JSON::ParserError => e
      render :layout => false, :json => { :err_message => '不正な操作です' } and return
    end
    flg_enable = (f_params[:flg_enable] == 'enable' ? true : false)
    SidfmdLocalVul.transaction do
      begin 
        SidfmdLocalVul.where(:id => ids).update_all(:flg_enable => flg_enable)
      rescue Exception => e
        logger.info(e)
        render :layout => false, :json => { :err_message => '変更に失敗しました' } and return
      end
    end
    Rails.cache.clear
    render :layout => false, :json => { :err_message => '' }
  end

  #一括削除
  def bulk_delete
    #f_params = params.fetch(:form, {}).permit(:ids)
    f_params = params.permit(:ids)
    err_message = ''
    begin 
      ids = JSON.parse(f_params[:ids])
    rescue JSON::ParserError => e
      err_message = '不正な操作です'
      render :layout => false, :json => { :err_message => err_message } and return
      #redirect_to :action => :index, :notice => '不正な操作です'
    end
   
    host_ids = []
    html_ids = []
    begin
      vuls = SidfmdLocalVul.where(:id => ids)
      html_ids = vuls.pluck(:html_id)
      host_ids = SidfmdVulView.relate_hostids(html_ids)
      vuls.destroy_all
      issue_ids = SidfmdIssue.where(:html_id => html_ids).pluck(:issue_id)
      if issue_ids.any? then
        SidfmdIssueCtag.where(:sidfmd_issue_id => issue_ids).delete_all
        Issue.where(:id => issue_ids).destroy_all
      end
    rescue Exception => e
      logger.info(e)
      err_message = '削除に失敗しました'
      render :layout => false, :json => { :err_message => err_message } and return
      #redirect_to :action => :index, :notice => '削除に失敗しました' and return
    end
    r = {
      :err_message => err_message,
      :host_ids => host_ids,
      :html_ids => html_ids
    }
    Rails.cache.clear
    render :layout => false, :json => r
    #redirect_to :action => :index
  end

  CSV_BOM = "\xff\xfe".force_encoding('UTF-16LE').freeze
  #export
  def export
    f_params = params.permit(:checked, :process, :executed_size)
    process = f_params[:process]
    err = false
    message = ''
    r = {}
    if process == 'start_export' then
      target_data = []
      begin
        checked = JSON.parse(f_params[:checked])
      rescue JSON::ParserError => e
        log_exception(e)
        err = true
        message = '不正な操作です。'
        render :layout => false, :json => r and return 
      end
      #if checked.any? then
        target_data = SidfmdLocalVul.where(:id => checked).sort_by{|vul| vul.title.downcase}
      #else
      #  target_data = SidfmdLocalVul.all.sort_by{|vul| vul.title.downcase}
      #end
      clear_export_session
      session[:export_local_vuls] = target_data
      r[:total_size] = target_data.size
      r[:err] = err
      r[:message] = message
      render :layout => false, :json => r
    elsif process == 'progress_export' then
      data = session[:export_local_vuls]
      if !data then
        err = true
        message = 'データの取得に失敗しました。'
      else
        begin
          target_data = data.shift(50)
          executed_size = target_data.size
          unless session[:export_local_vul_lines] then
            #vul-infoファイル・cveファイル・productファイル・更新履歴(changelog)ファイル用にデータを作成する
            session[:export_local_vul_lines] = {}
            session[:export_local_vul_lines][:vul_info] = []
            session[:export_local_vul_lines][:cve] = []
            session[:export_local_vul_lines][:product] = []
            session[:export_local_vul_lines][:change_log] = []
          end
          file_lines = session[:export_local_vul_lines]
          #クエリの発行を減らすために最初にデータをまとめる
          areas = SidfmdArea.all.map{|area| [ area.area_id, area.name ] }.to_h
          risks = SidfmdRisk.all.map{|risk| [ risk.risk_id, risk.name ] }.to_h
          sris = {
            SidfmdIssueSri::VALUE_LOW => SidfmdIssueSri::RISK_LOW ,
            SidfmdIssueSri::VALUE_MEDIUM => SidfmdIssueSri::RISK_MEDIUM,
            SidfmdIssueSri::VALUE_HIGH => SidfmdIssueSri::RISK_HIGH,
            SidfmdIssueSri::VALUE_CRITICAL => SidfmdIssueSri::RISK_CRITICAL
          }
          categories = SidfmdCategory.where(:is_system => true).map{|cat| [ cat.id, cat.name] }.to_h
          vul_ids = target_data.map{|data| data['id']}
          target_vul_sris = {}
          target_vul_cves = {}
          vul_change_logs = {}
          vul_products = {}
          SidfmdLocalVulSri.where(:sidfmd_local_vul_id => vul_ids).each do |sri| 
            target_vul_sris[sri.sidfmd_local_vul_id] = [] unless target_vul_sris[sri.sidfmd_local_vul_id]
            target_vul_sris[sri.sidfmd_local_vul_id] << sri
          end
          SidfmdLocalVulCve.where(:sidfmd_local_vul_id => vul_ids).each do |cve|
            target_vul_cves[cve.sidfmd_local_vul_id] = [] unless target_vul_cves[cve.sidfmd_local_vul_id]
            target_vul_cves[cve.sidfmd_local_vul_id] << cve 
          end
          SidfmdLocalVulLog.where(:sidfmd_local_vul_id => vul_ids).order(:id).each do |log|
            vul_change_logs[log.sidfmd_local_vul_id] = [] unless vul_change_logs[log.sidfmd_local_vul_id]
            vul_change_logs[log.sidfmd_local_vul_id] << log
          end
          #SIDプロダクト
          local_vul_app = SidfmdLocalVulApp.joins('LEFT OUTER JOIN sidfmd_products ON sidfmd_products.pid = sidfmd_local_vul_apps.pid')
                                           .joins('LEFT OUTER JOIN sidfmd_versions ON sidfmd_versions.vid = sidfmd_local_vul_apps.vid')
                                           .where('sidfmd_local_vul_apps.sidfmd_local_vul_id IN (?)', vul_ids)
                                           .select(
                                             'sidfmd_local_vul_apps.sidfmd_local_vul_id AS vul_id',
                                             'sidfmd_products.name AS p_name',
                                             'sidfmd_versions.fullname AS v_name'
                                           )
          local_vul_app.each do |app|
            vul_products[app.vul_id] = [] unless vul_products[app.vul_id]
            vul_products[app.vul_id] << {
              :vul_id => app.vul_id,
              :kind => 'manual_product',
              :name => (app.v_name ? app.v_name : app.p_name)
            }
          end
          #ローカルプロダクト
          local_vul_lapp = SidfmdLocalVulLapp.joins(
            'LEFT OUTER JOIN sidfmd_lproducts ON sidfmd_lproducts.id = sidfmd_local_vul_lapps.sidfmd_lproduct_id'
          )
          .joins(
            'LEFT OUTER JOIN sidfmd_lversions ON sidfmd_lversions.id = sidfmd_local_vul_lapps.sidfmd_lversion_id'
          )
          .where('sidfmd_local_vul_lapps.sidfmd_local_vul_id IN (?)', vul_ids)
          .select(
            'sidfmd_local_vul_lapps.sidfmd_local_vul_id AS vul_id',
            'sidfmd_lproducts.name AS p_name',
            'sidfmd_lversions.fullname AS v_name'
          )
          local_vul_lapp.each do |lapp|
            vul_products[lapp.vul_id] = [] unless vul_products[lapp.vul_id]
            vul_products[lapp.vul_id] << {
              :vul_id => lapp.vul_id,
              :kind => 'local_product',
              :name => (lapp.v_name ? lapp.v_name : lapp.p_name)            
            }
          end
          #ファイルのデータ生成
          target_data.each do |vul|
            vul = vul.symbolize_keys
            vul_sris = target_vul_sris[vul[:id]]
            #vul_info
            vul_data = {
              :id => vul[:id],
              :title => vul[:title],
              :area => areas[vul[:area_id]],
              :risk => risks[vul[:risk_id]],
              :short_description => vul[:short_description],
              :description => vul[:description],
              :flg_enable => vul[:flg_enable],
              :confirmation => vul[:confirmation],
              :workaround => vul[:workaround],
              :related => vul[:related],
              :category_risks => vul_sris.map{|sri| "#{categories[sri.category_id]}: #{sris[sri.risk_id]}" }.join(", "),#cat_name: risk_name
              :sri_memo => vul_sris.first.memo
            }
            file_lines[:vul_info] << vul_data
            #cve
            cves = target_vul_cves[vul[:id]]
            if cves then
              cves = cves.sort_by{|cve| cve.cve_id}
              cves.each do |cve|
                file_lines[:cve] << { 
                  :vul_id => cve.sidfmd_local_vul_id,
                  :cve_id => cve.cve_id,
                  :kind => 'cve',
                  :cvss_score => cve.c3_base_score,
                  :exploitable => cve.c3_exploitable,
                  :vector => cve.c3_vector
                }
              end
            end
            #product
            vul_prods = vul_products[vul[:id]]
            if vul_prods then
              vul_prods = vul_prods.sort{|a, b|
                ret = (a[:kind] == 'manual_product' ? 100 : 1000) <=> (b[:kind] == 'manual_product' ? 100 : 1000)
                if ret == 0 then
                  ret = a[:name] <=> b[:name]
                end
                ret
              }
              file_lines[:product].push(*vul_prods)
            end
            #change_log
            vul_logs = vul_change_logs[vul[:id]]
            if vul_logs then
              num = 1
              vul_logs.each do |log|
                vul_log = {
                  :vul_id => log.sidfmd_local_vul_id,
                  :kind => 'change_log',
                  :log => log.mes,
                  :order => num
                }
                num += 1
                file_lines[:change_log] << vul_log
              end
            end
          end
          session[:export_local_vul_lines] = file_lines
          r[:executed_size] = executed_size
        rescue Exception => e
          log_exception(e)
          err = true
          message = 'データの処理に失敗しました。'
        end
      end
      r[:err] = err
      r[:message] = message
      render :layout => false, :json => r
    else
      vuls_info_header = [
        '脆弱性登録ID'.encode('UTF-16LE'),
        '脆弱性ID'.encode('UTF-16LE'),
        '脆弱性名'.encode('UTF-16LE'),
        '影響範囲'.encode('UTF-16LE'),
        '危険度'.encode('UTF-16LE'),
        '簡易説明'.encode('UTF-16LE'),
        '概要'.encode('UTF-16LE'),
        '有効無効'.encode('UTF-16LE'),
        '影響などの確認方法'.encode('UTF-16LE'),
        '対処方法'.encode('UTF-16LE'),
        '関連情報'.encode('UTF-16LE'),
        'システムカテゴリ別のリスクの影響度'.encode('UTF-16LE'),
        'SRI補足情報'.encode('UTF-16LE'),
        '削除'.encode('UTF-16LE')
      ]
      cve_header = [
        '脆弱性ID'.encode('UTF-16LE'),
        '種類'.encode('UTF-16LE'),
        'CVSSスコア'.encode('UTF-16LE'),
        'CVEID'.encode('UTF-16LE'),
        '攻撃コードの有無'.encode('UTF-16LE'),
        'CVSS指標'.encode('UTF-16LE')
      ]
      product_header = [
        '脆弱性ID'.encode('UTF-16LE'),
        '種類'.encode('UTF-16LE'),
        'プロダクト名'.encode('UTF-16LE')
      ]
      changelog_header = [
        '脆弱性ID'.encode('UTF-16LE'),
        '種類'.encode('UTF-16LE'),
        '更新履歴'.encode('UTF-16LE'),
        '順位'.encode('UTF-16LE')
      ]

      new_cve_header = [
        '脆弱性登録ID'.encode('UTF-16LE'),
        '種類'.encode('UTF-16LE'),
        'CVSSスコア'.encode('UTF-16LE'),
        'CVEID'.encode('UTF-16LE'),
        '攻撃コードの有無'.encode('UTF-16LE'),
        'CVSS指標'.encode('UTF-16LE') 
      ]
      new_product_header = [
        '脆弱性登録ID'.encode('UTF-16LE'),
        '種類'.encode('UTF-16LE'),
        'プロダクト名'.encode('UTF-16LE')
      ]
      new_changelog_header = [
        '脆弱性登録ID'.encode('UTF-16LE'),
        '種類'.encode('UTF-16LE'),
        '更新履歴'.encode('UTF-16LE'),
        '順位'.encode('UTF-16LE')
      ]

      file_lines = session[:export_local_vul_lines]
      vul_info_lines = file_lines['vul_info']
      cve_lines = file_lines['cve']
      product_lines = file_lines['product']
      log_lines = file_lines['change_log']
      Tempfile.create('temp') do |f|
        path = f.path
        Zip::File.open(path, Zip::File::CREATE) do |zipfile|
          zipfile.get_output_stream("local-vulnerability-info.csv") do |f|
            f.puts(
              CSV.generate("#{CSV_BOM}", :col_sep => "\t".encode('UTF-16LE')) do |o|
                o << vuls_info_header
                vul_info_lines.each do |vul|
                  vul = vul.symbolize_keys
                  o << [
                    ''.encode('UTF-16LE'),
                    vul[:id].to_s.encode('UTF-16LE'),
                    vul[:title].encode('UTF-16LE'),
                    vul[:area].encode('UTF-16LE'),
                    vul[:risk].encode('UTF-16LE'),
                    vul[:short_description].encode('UTF-16LE'),
                    vul[:description].encode('UTF-16LE'),
                    vul[:flg_enable].to_s.encode('UTF-16LE'),
                    vul[:confirmation].encode('UTF-16LE'),
                    vul[:workaround].encode('UTF-16LE'),
                    vul[:related].encode('UTF-16LE'),
                    vul[:category_risks].encode('UTF-16LE'),
                    vul[:sri_memo].encode('UTF-16LE'),
                    ''.encode('UTF-16LE'),
                  ]
                end
              end
            )
          end
          zipfile.get_output_stream("opt/cve.csv") do |f|
            f.puts(
              CSV.generate("#{CSV_BOM}", :col_sep => "\t".encode('UTF-16LE')) do |o|
                o << cve_header
                cve_lines.each do |cve|
                  cve = cve.symbolize_keys
                  o << [
                    cve[:vul_id].to_s.encode('UTF-16LE'),
                    cve[:kind].encode('UTF-16LE'),
                    (cve[:cvss_score] ? cve[:cvss_score].to_s : '').encode('UTF-16LE'),
                    cve[:cve_id].encode('UTF-16LE'),
                    cve[:exploitable].to_s.encode('UTF-16LE'),
                    (cve[:vector] ? cve[:vector] : '').encode('UTF-16LE')
                  ]
                end
              end
            )
          end 
          zipfile.get_output_stream("opt/product.csv") do |f|
            f.puts(
              CSV.generate("#{CSV_BOM}", :col_sep => "\t".encode('UTF-16LE')) do |o|
                o << product_header
                product_lines.each do |prod|
                  prod = prod.symbolize_keys
                  o << [
                    prod[:vul_id].to_s.encode('UTF-16LE'),
                    prod[:kind].encode('UTF-16LE'),
                    prod[:name].encode('UTF-16LE') 
                  ]
                end
              end
            )
          end
          zipfile.get_output_stream("opt/changelog.csv") do |f|
            f.puts(
              CSV.generate("#{CSV_BOM}", :col_sep => "\t".encode('UTF-16LE')) do |o|  
                o << changelog_header
                log_lines.each do |log|
                  o << [
                    log[:vul_id].to_s.encode('UTF-16LE'),
                    log[:kind].encode('UTF-16LE'),
                    log[:log].encode('UTF-16LE'),
                    log[:order].to_s.encode('UTF-16LE') 
                  ] 
                end
              end
            )
          end
          #新規登録用
          zipfile.get_output_stream("regist/cve.csv") do |f|
            f.puts(
              CSV.generate("#{CSV_BOM}", :col_sep => "\t".encode('UTF-16LE')) do |o|
                o << new_cve_header
              end
            )
          end
          zipfile.get_output_stream("regist/product.csv") do |f|
            f.puts(
              CSV.generate("#{CSV_BOM}", :col_sep => "\t".encode('UTF-16LE')) do |o|
                o << new_product_header
              end 
            )
          end
          zipfile.get_output_stream("regist/changelog.csv") do |f|
            f.puts(
              CSV.generate("#{CSV_BOM}", :col_sep => "\t".encode('UTF-16LE')) do |o|
                o << new_changelog_header
              end
            )
          end
        end
        send_data(File.read(path), :filename => "vm-local-vulnerability-#{Time.now.strftime('%Y%m%d%H%M%S')}.zip", :disposition => 'attachment')
        cookies['export_flg'] = true
        clear_export_session
      end
    end

  end

  def read_import_file
    f_params = params.permit(:import_file)
    zip_file = f_params[:import_file]
    err = false
    unless zip_file then
      logger.info('invalid zip file')
      render :layout => false, :json => {:err => err} and return
    end
    
    clear_import_session
    file_data = {}
    file_data[:vul_info] = nil
    #既存データ用のhash
    file_data[:opt] = {}
    file_data[:opt][:cve] = {}
    file_data[:opt][:product] = {}
    file_data[:opt][:change_log] = {}
    #新規データ用のhash
    file_data[:regist] = {}
    file_data[:regist][:cve] = {}
    file_data[:regist][:product]= {}
    file_data[:regist][:change_log] = {}
    encode_file = ->(file){
      if file.getbyte(0) == 255 && file.getbyte(1) == 254 then
        file = file[2, file.length - 2]
      end
      file = file.force_encoding('UTF-16LE').encode('UTF-8')
    }
    vul_info_lines = []
    session[:import_notices] = []
    begin
      Zip::File.open(zip_file) do |zip|
        zip.each do |entry|
          next if entry.name && (entry.name.index('_MACOSX') || entry.name.index('.DS_Store')) #Mac OSで入ってくるファイル
          if entry.file? then
            if entry.name && entry.name.index('local-vulnerability-info.csv') then
              file_data[:vul_info] = { :file_name => entry.name, :file => encode_file.(entry.get_input_stream.read) }
            elsif entry.name && entry.name.index('opt/') then
              if entry.name.index('cve.csv') then
                file_data[:opt][:cve] = { :file_name => entry.name, :file => encode_file.(entry.get_input_stream.read) } 
              elsif entry.name.index('product.csv') then
                file_data[:opt][:product] = { :file_name => entry.name, :file => encode_file.(entry.get_input_stream.read) }
              elsif entry.name.index('changelog.csv') then
                file_data[:opt][:change_log] = { :file_name => entry.name, :file => encode_file.(entry.get_input_stream.read) }
              end
            elsif entry.name && entry.name.index('regist/') then
              if entry.name.index('cve.csv') then
                file_data[:regist][:cve] = { :file_name => entry.name, :file => encode_file.(entry.get_input_stream.read) }
              elsif entry.name.index('product.csv') then
                file_data[:regist][:product] = { :file_name => entry.name, :file => encode_file.(entry.get_input_stream.read) }
              elsif entry.name.index('changelog.csv') then
                file_data[:regist][:change_log] = { :file_name => entry.name, :file => encode_file.(entry.get_input_stream.read) }
              end
            end
          end
        end
      end
     if !file_data[:vul_info] then
       logger.info('local-vulnerability-info.csv is not found')
       err = true
       session[:import_notices] << 'local-vulnerability-info.csvファイルがありません。'
     else
       vul_info_line_size = CSV.parse(file_data[:vul_info][:file], :col_sep => "\t", :headers => :first_row).size
     end  
    rescue Exception => e
      log_exception(e)
      err = true
    end

    session[:import_actions] = []
    session[:import_local_vul_files] = file_data

    render :layout => false, :json => { :err => err, :total_size => vul_info_line_size }
  end

  def progress_import
    files = session[:import_local_vul_files]

    err = false 
    if !files['vul_info']['lines'] then
      #最初の実行時に読み取ったファイルからデータを生成
      file_data = create_data_from_files(files)
      files = file_data[:files]
      err = file_data[:err]
      if err then
        render :layout => false, :json => { :err => err } and return
      end
    end

    #vul_infoファイルのデータ
    target_vul_lines = files['vul_info']['lines'].shift(30)
    vul_info_file_name = files['vul_info']['file_name']

    #optフォルダ下のファイルのデータ
    if files['opt']['cve'].any? then
      opt_cve_lines = files['opt']['cve']['lines']
      opt_cve_file_name = files['opt']['cve']['file_name']
    else
      opt_cve_lines = {}
    end

    if files['opt']['product'].any? then
      opt_prod_lines = files['opt']['product']['lines']
      opt_prod_file_name = files['opt']['product']['file_name']
    else
      opt_prod_lines = {}
    end

    if files['opt']['change_log'].any? then
      opt_log_lines = files['opt']['change_log']['lines']
      opt_log_file_name = files['opt']['change_log']['file_name']
    else
      opt_prod_lines = {}
    end

    #registフォルダ下のファイルのデータ
    if files['regist']['cve'].any? then
      regist_cve_lines = files['regist']['cve']['lines']
      regist_cve_file_name = files['regist']['cve']['file_name']
    else
      regist_cve_lines = {}
    end

    if files['regist']['product'].any? then
      regist_prod_lines = files['regist']['product']['lines']
      regist_prod_file_name = files['regist']['product']['file_name']
    else
      regist_prod_lines = {}
    end

    if files['regist']['change_log'].any? then
      regist_log_lines = files['regist']['change_log']['lines']
      regist_log_file_name = files['regist']['change_log']['file_name']  
    else
      regist_log_lines = {}
    end

    #実際にimportするデータ(インポート確認画面で表示して登録するデータ)を生成
    areas = SidfmdArea.all.map{|area| [ area.area_id, area.name ] }.to_h
    risks = SidfmdRisk.all.map{|risk| [ risk.risk_id, risk.name ] }.to_h
    sris = {
      SidfmdIssueSri::VALUE_LOW => SidfmdIssueSri::RISK_LOW,
      SidfmdIssueSri::VALUE_MEDIUM => SidfmdIssueSri::RISK_MEDIUM,
      SidfmdIssueSri::VALUE_HIGH => SidfmdIssueSri::RISK_HIGH,
      SidfmdIssueSri::VALUE_CRITICAL => SidfmdIssueSri::RISK_CRITICAL
    }
    categories = SidfmdCategory.where(:is_system => true).map{|cat| [ cat.id, cat.name ] }.to_h

    actions = session[:import_actions]
    executed_size = target_vul_lines.size

    target_vul_lines.each do |line|
      line = line.symbolize_keys
      if (line[:new_vul_id] && line[:new_vul_id].size > 0) || ((!line[:new_vul_id] || line[:new_vul_id].size == 0) && (!line[:vul_id] || line[:vul_id].size == 0)) then
        #新規登録
        cves = (regist_cve_lines ? (regist_cve_lines[line[:new_vul_id]] || regist_cve_lines[line[:new_vul_id].to_s]) : nil)
        prods = (regist_prod_lines ? (regist_prod_lines[line[:new_vul_id]] || regist_prod_lines[line[:new_vul_id].to_s]) : nil)
        logs = (regist_log_lines ? (regist_log_lines[line[:new_vul_id]] || regist_log_lines[line[:new_vul_id].to_s]): nil)  
        category_risks = []
        line[:category_risks].split(',').each do |cat_risk|
          cat_risk = cat_risk.split(':')
          cat_name = cat_risk[0].strip
          risk_name = cat_risk[1].strip
          category_risks << [ cat_name, risk_name ]
        end
        line[:category_risks] = category_risks.to_h
        action = {
          :action => 'create',
          :vul_title => line[:title],
          :vul => line,
          :action_name => '新規登録'
        }
        #registデータが正しい場合は登録、不正な場合は登録しない
        if cves then
          filtered_cves_data = filter_cves(cves, regist_cve_file_name)
          is_pass = filtered_cves_data[:is_pass]
          if is_pass then
            filtered_cves_data[:cves].each do |cve|
              if cve[:vector] && cve[:vector].size > 0 then
                cve[:vector] = cve[:vector].split('/').map{|v| v.split(':')}.to_h
              end
            end
            action[:cves] = filtered_cves_data[:cves]
          end
        end
        if prods then
          filtered_prods = filter_prods(prods, regist_prod_file_name)
          action[:products] = filtered_prods
        end
        if logs then
          filtered_logs = filter_logs(logs, regist_log_file_name, true)
          action[:logs] = filtered_logs
        end
        actions << action
      else
        #既存
        vul = SidfmdLocalVul.find_by(:id => line[:vul_id])
        if !vul then
          session[:import_notices] << "#{vul_info_file_name}の#{line[:line_num]}行目: 脆弱性IDに誤りがあるか、脆弱性が削除されています。"
          next          
        end
        cves = (opt_cve_lines ? (opt_cve_lines[line[:vul_id]] || opt_cve_lines[line[:vul_id].to_s]): nil)
        prods = (opt_prod_lines ? (opt_prod_lines[line[:vul_id]] || opt_prod_lines[line[:vul_id].to_s]) : nil)
        logs = (opt_log_lines ? (opt_log_lines[line[:vul_id]] || opt_log_lines[line[:vul_id].to_s]): nil)
     
        vul_cve = nil
        vul_prods = nil
        vul_logs = nil
        if line[:flg_delete] && line[:flg_delete].size > 0 && line[:flg_delete] == 'true' then
          #削除
          action = {
            :action => 'delete',
            :vul_title => line[:title],
            :vul_id => vul.id,
            :action_name => '削除'
          }
          actions << action
          next
        end
 
        #変更チェック
        #※optデータが不正の場合は削除扱い
        options = []
        if cves then
          filtered_cves_data = filter_cves(cves, opt_cve_file_name)
          if filtered_cves_data[:is_pass] then
            filtered_cves_data[:cves].each do |cve|
              if cve[:vector] && cve[:vector].size > 0 then
                cve[:vector] = cve[:vector].split('/').map{|v| v.split(':')}.to_h
              end
            end
            vul_cves = filtered_cves_data[:cves]
          end
        end
        if prods then
          filtered_prods = filter_prods(prods, opt_prod_file_name)
          vul_prods = filtered_prods
        end
        if logs then
          filtered_logs = filter_logs(logs, opt_log_file_name, false)  
          vul_logs = filtered_logs
        end
 
        #基本データ(vul_info)比較
        line[:short_description] = '' unless line[:short_description]
        line[:description] = '' unless line[:description]
        line[:confirmation] = '' unless line[:confirmation]
        line[:workaround] = '' unless line[:workaround]
        line[:related] = '' unless line[:related]
        line[:sri_memo] = '' unless line[:sri_memo] 
        if vul.title != line[:title] then
          options << create_vul_info_option(vul.title, line[:title], 'title_change', '脆弱性名変更')
        end
        if areas[vul.area_id] != line[:area] then
          options << create_vul_info_option(areas[vul.area_id], line[:area], 'area_change', '影響範囲変更')
        end
        if risks[vul.risk_id] != line[:risk] then
          options << create_vul_info_option(risks[vul.risk_id], line[:risk], 'risk_change', '危険度変更')
        end
        if vul.short_description != line[:short_description] then
          options << create_vul_info_option(vul.short_description, line[:short_description], 'short_description_change', '簡易説明変更')
        end
        if vul.description != line[:description] then
          options << create_vul_info_option(vul.description, line[:description], 'description_change', '概要変更')
        end
        if vul.flg_enable.to_s != line[:flg_enable] then
          options << create_vul_info_option(vul.flg_enable.to_s, line[:flg_enable], 'flg_enable_change', '有効無効変更')
        end
        if vul.confirmation != line[:confirmation] then
          options << create_vul_info_option(vul.confirmation, line[:confirmation], 'confirmation_change', '影響などの確認方法変更')
        end
        if vul.workaround != line[:workaround] then
          options << create_vul_info_option(vul.workaround, line[:workaround], 'workaround_change', '対処方法変更')
        end
        if vul.related != line[:related] then
          options << create_vul_info_option(vul.related, line[:related], 'related_change', '関連情報変更')
        end
        vul_category_risks = {}
        vul_sri_memo = nil
        vul.sidfmd_local_vul_sris.each do |vul_sri|
          sri_name = sris[vul_sri.risk_id]
          category_name = categories[vul_sri.category_id]
          vul_category_risks[category_name] = sri_name
          vul_sri_memo = vul_sri.memo unless vul_sri_memo
        end
        is_category_risk_change = false
        new_category_risks = []
        category_risks = line[:category_risks].split(',')
        category_risks.each do |cat_risk|
          cat_risk = cat_risk.split(':')
          cat_name = cat_risk[0].strip
          risk_name = cat_risk[1].strip
          if vul_category_risks[cat_name] != risk_name then
            is_category_risk_change = true
          end
          new_category_risks << [ cat_name, risk_name ]
        end
        if is_category_risk_change then
          options << create_vul_info_option(vul_category_risks, new_category_risks.to_h, 'category_risks_change', 'システムカテゴリ別のリスクの影響度変更')
        end
        if vul_sri_memo != line[:sri_memo] then
          options << create_vul_info_option(vul_sri_memo, line[:sri_memo], 'sri_memo_change', 'SRI補足情報変更')
        end

        #product比較
        apps = vul.sidfmd_local_vul_apps
        lapps = vul.sidfmd_local_vul_lapps
        old_app_pids = []
        old_app_vids = []
        apps.each do |app|
          if !app.vid || app.vid == 0 then
            old_app_pids << app.pid
          else
            old_app_vids << app.vid
          end
        end
        old_app_prds = SidfmdProduct.where(:pid => old_app_pids)
        old_app_vers = SidfmdVersion.where(:vid => old_app_vids)
        old_app_prd_names = old_app_prds.pluck(:name)
        old_app_ver_names = old_app_vers.pluck(:fullname)
        old_prd_names = old_app_prd_names + old_app_ver_names
        
        old_lapp_pids = []
        old_lapp_vids = []
        lapps.each do |lapp|
          if !lapp.sidfmd_lversion_id || lapp.sidfmd_lversion_id == 0 then
            old_lapp_pids << lapp.sidfmd_lproduct_id
          else
            old_lapp_vids << lapp.sidfmd_lversion_id
          end
        end
        old_lapp_prds = SidfmdLproduct.where(:id => old_lapp_pids)
        old_lapp_vers = SidfmdLversion.where(:id => old_lapp_vids)
        old_lapp_prd_names = old_lapp_prds.pluck(:name)
        old_lapp_ver_names = old_lapp_vers.pluck(:fullname) 
        old_prd_names = (old_prd_names + old_lapp_prd_names + old_lapp_ver_names)
 
        new_prd_names = []
        if vul_prods then
          new_prd_names = vul_prods.map{|prd| prd[:name]}
        end
        if (old_prd_names - new_prd_names).any? || (new_prd_names - old_prd_names).any? then
          option = create_product_option(old_app_prds, old_app_vers, old_lapp_prds, old_lapp_vers, vul_prods)
          options << option
        end

        #cve比較
        current_cves = vul.sidfmd_local_vul_cves
        current_cve_size = current_cves.count
        new_cve_size = vul_cves ? vul_cves.size : 0
        create_current_cves_data = -> (){
          current_cves_data = {}
          current_cves.each do |cve|
            current_cves_data[cve.cve_id] = {
              :cve_id => cve.cve_id,
              :cvss_score => cve.c3_base_score,
              :vector => cve.c3_vector ? cve.c3_vector.split('/').map{|v| v.split(':')}.to_h : nil,
              :exploitable => cve.c3_exploitable.to_s.downcase
            }
          end
          current_cves_data
        }
        if current_cve_size != new_cve_size then
          #[ 'CVSSスコア', 'CVE ID', '攻撃コードの有無', 'CVSS指標' ].each do |item_name|
          #  options << create_cve_option(current_cves, vul_cve, item_name)
          #end
          options << create_cve_option(create_current_cves_data.(), vul_cves)
        elsif current_cve_size > 0 && vul_cves then
          current_cves_data = create_current_cves_data.()
          vul_cves.each do |cve|
            current_cve = current_cves_data[cve[:cve_id]]
            if !current_cve ||
               current_cve[:cvss_score] != cve[:cvss_score] ||
               current_cve[:exploitable] != cve[:exploitable].to_s.downcase ||
               current_cve[:vector] != cve[:vector]
            then
              options << create_cve_option(current_cves_data, vul_cves)
              break
            end
          end
          #first_current_cve = current_cves.first
          #current_cve_ids = current_cves.map{|cve| cve.cve_id}.sort.join(',')
          #if first_current_cve.c3_base_score != vul_cve[:cvss_score] then
          #  options << create_cve_option(current_cves, vul_cve, 'CVSSスコア')
          #end
          #if current_cve_ids != vul_cve[:cve_ids] then
          #  options << create_cve_option(current_cves, vul_cve, 'CVE ID')
          #end
          #if first_current_cve.c3_exploitable.to_s != vul_cve[:exploitable] then
          #  options << create_cve_option(current_cves, vul_cve, '攻撃コードの有無')
          #end
          #if first_current_cve.c3_vector != vul_cve[:vector] then
          #  options << create_cve_option(current_cves, vul_cve, 'CVSS指標')
          #end
        end
  
        #更新履歴比較
        current_logs = vul.sidfmd_local_vul_logs.order(:id).map{|log| {:log => log.mes} }
        if (current_logs.any? && !vul_logs) || (current_logs.empty? && vul_logs) then
          options << create_log_option(current_logs, vul_logs)
        elsif current_logs.any? && vul_logs then
          if current_logs.size != vul_logs.size then
            options << create_log_option(current_logs, vul_logs)
          else
            for idx in 0..(current_logs.size - 1) do
              old_log = current_logs[idx][:log]
              new_log = vul_logs[idx][:log]
              if old_log != new_log then
                options << create_log_option(current_logs, vul_logs)
                break
              end
            end
          end      
        end
        if options.size > 0 then
          action = {
            :action => 'modify',
            :vul_title => vul.title,
            :vul_id => vul.id,
            :options => options,
            :action_name => '変更'
          }
          actions << action
        end

      end      
    end

    session[:import_local_vul_files] = files
    session[:import_actions] = actions
    render :layout => false, :json => { :executed_size => executed_size, :err => err }
  end

  def import
    @notices = session[:import_notices]
    @actions = session[:import_actions]
    if !@actions then
      redirect_to :action => :index, :notice => 'インポートに失敗しました。' and return
    elsif @actions.empty? then
      redirect_to :action => :index, :notice => '変更がありません。' and return
    end
    @actions = @actions.sort{|a, b|
      r = a['action'] <=> b['action']
      if r == 0 then
        r = a['vul_title'] <=> b['vul_title']
      end
      r
    }
  end

  def import_apply
    f_params = params.permit(:action_data)
    err_message = ''
    r = {}
    action = JSON.parse(f_params[:action_data], :symbolize_names => true)
    case action[:action]
    when 'create' then
      #新規登録
      areas = SidfmdArea.all.map{|area| [ area.name, area.area_id ] }.to_h
      risks = SidfmdRisk.all.map{|risk| [ risk.name, risk.risk_id ] }.to_h
      sris = {
        SidfmdIssueSri::RISK_LOW => SidfmdIssueSri::VALUE_LOW,
        SidfmdIssueSri::RISK_MEDIUM => SidfmdIssueSri::VALUE_MEDIUM,
        SidfmdIssueSri::RISK_HIGH => SidfmdIssueSri::VALUE_HIGH,
        SidfmdIssueSri::RISK_CRITICAL => SidfmdIssueSri::VALUE_CRITICAL
      }
      categories = SidfmdCategory.where(:is_system => true).map{|cat| [ cat.name, cat.id ] }.to_h
      
      new_vul = action[:vul]
      vul_title = action[:vul_title]
      begin 
        SidfmdLocalVul.transaction do
          area_id = areas[new_vul[:area]]
          risk_id = risks[new_vul[:risk]]
          if !area_id || !risk_id then
             logger.info('invalid area_id') unless area_id
             logger.info('invalid risk_id') unless risk_id
             raise Exception
          end
          vul = SidfmdLocalVul.new(
            flg_enable: new_vul[:flg_enable].to_s == 'true' ? true : false,
            title: new_vul[:title],
            risk_id: risk_id,
            area_id: area_id,
            short_description: (new_vul[:short_description] ? new_vul[:short_description] : ''),
            description: (new_vul[:description] ? new_vul[:description] : ''),
            confirmation: (new_vul[:confirmation] ? new_vul[:confirmation] : ''),
            workaround: (new_vul[:workaround] ? new_vul[:workaround] : ''),
            related: (new_vul[:related] ? new_vul[:related] : ''),
          )
          if vul.save then
            vul_sris = []
            vul_sri_memo = (new_vul[:sri_memo] ? new_vul[:sri_memo] : '')
            new_vul[:category_risks].each do |cat_name, risk_name|
              cat_name = cat_name.to_s
              cat_id = categories[cat_name]
              risk_id = risks[risk_name]
              if !cat_id || !risk_id then
                logger.info('invalid category_id') unless cat_id
                logger.info("invalid category's risk_id") unless risk_id
                raise Exception
              end
              vul_sris << SidfmdLocalVulSri.new(
                sidfmd_local_vul_id: vul.id,
                category_id: cat_id,
                risk_id: risk_id,
                memo: vul_sri_memo
              )
            end
            if vul_sris.size != 8 then
              logger.info('invalid category_names')
              raise Exception
            end
            SidfmdLocalVulSri.import(vul_sris)
            #CVE
            if action[:cves] then
              import_cves = []
              action[:cves].each do |cve|
                vector = cve[:vector] ? cve[:vector].map{|k, v| "#{k.to_s}:#{v}"}.join('/') : nil
                if vector then
                  is_vector_correct = check_cve_vector(vector)
                  if !is_vector_correct then
                    logger.info('invalid cve_vector')
                    raise Exception
                  end
                end
                cvss_score = cve[:cvss_score] ? cve[:cvss_score].to_f : nil
                exploitable = cve[:exploitable].to_s == 'true'
                if(cve[:cve_id] =~ /^CVE-[0-9]{4}-[0-9]+$/) == nil then
                  logger.info('invalid cve_id')
                  raise Exception
                elsif cvss_score && (cvss_score > 10 || cvss_score < 0) then
                  logger.info('invalid cvss_score')
                  raise Exception
                end
                if (vector and cvss_score) or (!vector and !cvss_score) then
                  import_cves << SidfmdLocalVulCve.new(
                    sidfmd_local_vul_id: vul.id,
                    cve_id: cve[:cve_id],
                    c3_base_score: cvss_score,
                    c3_vector: vector,
                    c3_exploitable: exploitable
                  )
                end
              end
              SidfmdLocalVulCve.import(import_cves) if import_cves.any?
            end
            #プロダクト
            products = action[:products]
            if products then
              apps = []
              lapps = []
              products.each do |prd|
                prd_name = prd[:name]
                case prd[:kind]
                when 'manual_product' then
                  target_ver = SidfmdVersion.find_by(:fullname => prd_name)
                  if target_ver then
                    apps << SidfmdLocalVulApp.new(
                      sidfmd_local_vul_id: vul.id,
                      pid: target_ver.sidfmd_product.pid,
                      vid: target_ver.vid
                    )
                  else 
                    target_prod = SidfmdProduct.find_by(:name => prd_name)
                    if target_prod then
                      apps << SidfmdLocalVulApp.new(
                        sidfmd_local_vul_id: vul.id,
                        pid: target_prod.pid,
                        vid: 0
                      )
                    end
                  end
                when 'local_product' then
                  target_lver = SidfmdLversion.find_by(:fullname => prd_name) 
                  if target_lver then
                    lapps << SidfmdLocalVulLapp.new(
                      sidfmd_local_vul_id: vul.id,
                      sidfmd_lproduct_id: target_lver.sidfmd_lproduct_id,
                      sidfmd_lversion_id: target_lver.id
                    )
                  else
                    target_lprod = SidfmdLproduct.find_by(:name => prd_name)
                    if target_lprod then
                      lapps << SidfmdLocalVulLapp.new(
                        sidfmd_local_vul_id: vul.id,
                        sidfmd_lproduct_id: target_lprod.id,
                        sidfmd_lversion_id: 0
                      )
                    end
                  end
                end
              end
              SidfmdLocalVulApp.import(apps) if apps.any?
              SidfmdLocalVulLapp.import(lapps) if lapps.any?
            end
            #更新履歴
            logs = action[:logs]
            if logs then
              time_now = Time.now
              logs.each do |log|
                #順序を保証するために、1ずつcreateする。データはファイルの「順位」順位なっている
                SidfmdLocalVulLog.create(
                  sidfmd_local_vul_id: vul.id,
                  dt: time_now,
                  mes: log[:log]
                )
              end
            end
          end
        end
      rescue Exception => e
        log_exception(e)
        err_message = "「#{vul_title}」の登録に失敗しました。"
      end
    when 'modify' then
      begin 
        vul_id = action[:vul_id]
        vul = SidfmdLocalVul.find_by(:id => vul_id)
        if !vul then
          logger.info('local vul is not found')
          raise Exception
        end
        options = action[:options]
        vul_title = action[:vul_title]

        SidfmdLocalVul.transaction do
          options.each do |option|
            new_data = option[:new_data]
            case option[:action]
            when 'title_change' then
              vul.update(:title => new_data) if new_data.size > 0
            when 'area_change' then
              areas = SidfmdArea.all.map{|area| [ area.name, area.area_id ] }.to_h
              area_id = areas[new_data]
              vul.update(:area_id => area_id) if area_id
            when 'risk_change' then
              risks = SidfmdRisk.all.map{|risk| [ risk.name, risk.risk_id ] }.to_h
              risk_id = risks[new_data]
              vul.update(:risk_id => risk_id) if risk_id
            when 'short_description_change' then
              new_data = '' if new_data.nil?
              vul.update(:short_description => new_data)
            when 'description_change' then
              new_data = '' if new_data.nil?
              vul.update(:description => new_data)
            when 'flg_enable_change' then
              new_data = new_data == 'true' ? true : false
              vul.update(:flg_enable => new_data)
            when 'confirmation_change' then
              new_data = '' if new_data.nil?
              vul.update(:confirmation => new_data) 
            when 'workaround_change' then
              new_data = '' if new_data.nil?
              vul.update(:workaround => new_data)
            when 'related_change' then
              new_data = '' if new_data.nil?
              vul.update(:related => new_data)
            when 'category_risks_change' then
              categories = SidfmdCategory.where(:is_system => true).map{|cat| [ cat.name, cat.id ] }.to_h
              risks = SidfmdRisk.all.map{|risk| [ risk.name, risk.risk_id ] }.to_h
              category_risks = {}
              new_data.each do |cat_name, risk_name|
                cat_name = cat_name.to_s
                cat_id = categories[cat_name]
                risk_id = risks[risk_name]
                if !cat_id || !risk_id then
                  logger.info('invalid category_id') unless cat_id
                  logger.info("invalid category's risk_id") unless risk_id
                  raise Exception
                end
                category_risks[cat_id] = risk_id
              end
              vul_sris = vul.sidfmd_local_vul_sris
              category_risks.each do |cat_id, risk_id|
                vul_sris.where(:category_id => cat_id).first.update(:risk_id => risk_id)
              end
              vul.update(:updated_at => Time.now)
            when 'sri_memo_change' then
              new_data = '' if new_data.nil?
              vul.sidfmd_local_vul_sris.update_all(:memo => new_data)
              vul.update(:updated_at => Time.now)
            when 'product_change' then
              apps = vul.sidfmd_local_vul_apps
              lapps = vul.sidfmd_local_vul_lapps
              pids = []
              vids = []
              lpids = []
              lvids = []
              lvid_with_lpid = {}
              new_data = [] unless new_data
              new_data.each do |prd|
                if prd[:kind] == 'manual_product' then
                  ver = SidfmdVersion.find_by(:fullname => prd[:name])
                  if ver then
                    vids << ver.vid
                  else
                    prod = SidfmdProduct.find_by(:name => prd[:name])
                    pids << prod.pid if prod
                  end
                elsif prd[:kind] == 'local_product' then
                  lver = SidfmdLversion.find_by(:fullname => prd[:name])
                  if lver then
                    lvids << lver.id
                    lvid_with_lpid[lver.id] = lver.sidfmd_lproduct_id
                  else
                    lprod = SidfmdLproduct.find_by(:name => prd[:name])
                    lpids << lprod.id if lprod
                  end 
                end
              end
              delete_app_ids = []
              delete_lapp_ids = []
              if apps.any? then
                delete_apps = apps.where.not(:vid => vids).where.not(:vid => 0)
                delete_apps.each{|app| vids.delete(app.vid)}
                delete_apps.destroy_all
 
                delete_apps = apps.where.not(:pid => pids).where(:vid => 0)
                delete_apps.each{|app| pids.delete(app.pid)}
                delete_apps.destroy_all
              end
              if lapps.any? then
                delete_lapps = lapps.where.not(:sidfmd_lversion_id => lvids).where.not(:sidfmd_lversion_id => 0)
                delete_lapps.each{|lapp| lvids.delete(lapp.sidfmd_lversion_id)}
                delete_lapps.destroy_all
  
                delete_lapps = lapps.where.not(:sidfmd_lproduct_id => lpids).where(:sidfmd_lversion_id => 0)
                delete_lapps.each{|lapp| lpids.delete(lapp.sidfmd_lproduct_id)}
                delete_lapps.destroy_all
              end
              new_apps = []
              new_lapps = []
              pids.each do |pid|
                new_apps << SidfmdLocalVulApp.new(
                  sidfmd_local_vul_id: vul.id,
                  pid: pid,
                  vid: 0
                )
              end
              vids.each do |vid|
                ver = SidfmdVersion.find_by(:vid => vid)
                if ver then
                  pid = SidfmdProduct.find(ver.sidfmd_product_id).pid
                  new_apps << SidfmdLocalVulApp.new(
                    sidfmd_local_vul_id: vul.id,
                    pid: pid,
                    vid: vid
                  )
                end
              end
              SidfmdLocalVulApp.import(new_apps) if new_apps.any?
              
              lpids.each do |lpid|
                new_lapps << SidfmdLocalVulLapp.new(
                  sidfmd_local_vul_id: vul.id,
                  sidfmd_lproduct_id: lpid,
                  sidfmd_lversion_id: 0
                )
              end
              lvids.each do |lvid|
                lpid = lvid_with_lpid[lvid]
                new_lapps << SidfmdLocalVulLapp.new(
                  sidfmd_local_vul_id: vul.id,
                  sidfmd_lproduct_id: lpid,
                  sidfmd_lversion_id: lvid
                )
              end
              SidfmdLocalVulLapp.import(new_lapps) if new_lapps.any?
              vul.update(:updated_at => Time.now)
            when 'cve_ids_change' then
              vul.sidfmd_local_vul_cves.destroy_all
              if new_data.is_a?(Array) then
                new_data.each do |cve|
                  next unless cve[:cve_id] =~ /\ACVE\-[0-9]{4}\-[0-9]+\z/
                  regist_vectors = create_regist_cvss_vectors cve[:vector]
                  SidfmdLocalVulCve.create(
                    sidfmd_local_vul_id: vul.id,
                    cve_id: cve[:cve_id],
                    c3_base_score: (regist_vectors.any? ? cve[:cvss_score] : nil),
                    c3_vector: ((regist_vectors.any? and cve[:cvss_score]) ? regist_vectors.join('/') : nil),
                    c3_exploitable: cve[:exploitable].to_s == 'true'
                  )
                end
              end 
              vul.update(:updated_at => Time.now)
            when 'logs_change' then
              new_data = [] unless new_data
              current_logs = vul.sidfmd_local_vul_logs.order(:id)
              if new_data.empty? then
                current_logs.destroy_all
              else
                current_logs = current_logs.to_a
                current_logs_size = current_logs.size
                orders = []
                new_vul_logs = []
                new_data = new_data.sort_by{|data| data[:order]}
                max_order = new_data.last[:order].to_i
                new_data.each do |data|
                  order = data[:order].to_i
                  new_log = data[:log]
                  orders << order
                  if order <= current_logs_size then
                    #既存データにあるものは、その履歴をupdate
                    #更新履歴に変更がある場合は、記入されている順位番目に作成された更新履歴をupdateする
                    current_log = current_logs[order - 1]
                    old_log = current_log.mes
                    current_log.update(:mes => new_log) if new_log != old_log
                  elsif order > current_logs_size then
                    #既存に無いものは作成。orderが既存の履歴数を超える = 追加分 
                    new_vul_logs << SidfmdLocalVulLog.new(
                      sidfmd_local_vul_id: vul.id,
                      dt: Time.now,
                      mes: new_log
                    )
                  end
                end
                if max_order < current_logs_size then
                  extra_logs = current_logs.slice(max_order..current_logs_size)
                  extra_logs.each{|log| log.destroy }
                end
                cnt = 0
                check_and_delete_log = ->(order){ #ファイルのデータに記入がないものを削除
                  if (cnt + 1) != order then
                    current_logs[cnt].destroy if current_logs[cnt]
                    cnt += 1
                    check_and_delete_log.(order)
                  else
                    cnt += 1
                    return true
                  end
                }
                orders.each do |odr|
                  check_and_delete_log.(odr)
                end 
                SidfmdLocalVulLog.import(new_vul_logs) if new_vul_logs.any?
                vul.update(:updated_at => Time.now)
              end
            end
          end
        end 
      rescue Exception => e
        log_exception(e)
        err_message = "「#{vul_title}」の変更に失敗しました。"
      end
    when 'delete' then
      vul_id = action[:vul_id]
      vul_title = action[:vul_title]
      SidfmdLocalVul.transaction do
        begin
          vul = SidfmdLocalVul.find_by(:id => vul_id)
          if vul then
            html_id = vul.html_id
            target_host_ids = SidfmdVulView.relate_hostids([ html_id ])
            vul.destroy
            issue_ids = SidfmdIssue.where(:html_id => html_id).pluck(:issue_id)
            if issue_ids.any? then
              SidfmdIssueCtag.where(:sidfmd_issue_id => issue_ids).delete_all
              Issue.where(:id => issue_ids).destroy_all
            end
            target_host_ids |= SidfmdVulView.relate_hostids([ html_id ])
            SidfmdVulView.sync_data([ html_id ])
            SidfmdHostView.sync_data(target_host_ids)
          end
        rescue Exception => e
          logger.info(e)
          err_message = "「#{vul_title}」の変更に失敗しました。"
        end
      end
    end
    Rails.cache.clear
    r[:err_message] = err_message
    render :layout => false, :json => r
  end

  #画面操作中に、リンクの入力が正しいかチェック 
  def check_texts
    f_params = params.permit(:texts_data)
    data = JSON.parse(f_params[:texts_data])
    convert_texts = data.map{|d|
      SidfmdLocalVul.get_link_converted_text(d['text'], d['label'])
    }
    err_messages = convert_texts.map{|d| d[:err]}.compact
    render :layout => false, :json => err_messages
  end

  #脆弱性チケットに関連するローカル脆弱性のhtml_ids 削除・変更時にアラートを出す。
  def get_issue_related_vul_ids
    f_params = params.permit(:vul_ids)
    vul_ids = JSON.parse(f_params[:vul_ids])
    issue_related_vul_ids = SidfmdIssue.joins('LEFT OUTER JOIN sidfmd_local_vuls ON sidfmd_issues.html_id = sidfmd_local_vuls.html_id')
                                       .where('sidfmd_local_vuls.id IN (?)', vul_ids)
                                       .select('sidfmd_local_vuls.id AS vul_id')
                                       .map{|s| s.vul_id}.uniq
    render :layout => false, :json => issue_related_vul_ids
  end

  def api_get_display_summary_content
    GC.start
    f_params = params.fetch(:pages, {}).permit(:offset, :per_page, :order, :filters, :checked, :all_checked)

    parse_params
    @vuls[:per_page] = f_params[:per_page].to_i

    check_filters

    @all_vuls = SidfmdLocalVul.all
    filter_vuls
    sort_vuls

    #フィルタ後にデータ取得する
    @all_vul_ids = []
    @all_vul_names = {}
    @all_vuls.each do |vul|
      @all_vul_ids << vul.id
      @all_vul_names[vul.id] = vul.title
    end

    @all_vuls = @all_vuls.slice(@vuls[:offset]..(@vuls[:offset] + @vuls[:per_page] - 1))
    @all_vuls = [] unless @all_vuls
    @vuls[:content] = create_content

    ret = {
      :rows => @vuls[:content].map{|lvul|
        create_summary_lvul_html(
          lvul,
          {
            :checked =>  @vuls[:checked],
          }
        )
      },
      :total => @all_vul_ids.size,
      :all_ids => @all_vul_ids,
      :hidden_params => @all_vul_names,
      :sidfm_roles => @sidfm_roles,
      :user_cats => (@sidfm_roles.include?(:admin) ? nil : User.current.categories.pluck(:id))
    }

    render :layout => false, :json => ret
  end

end
